/*Author: Jin-Hee Cho 
* Project Name: Wireless Ad Hoc Messenger Using Pocket PCs
* Purpose: Thesis for Master Science in Computer Science 
* at Virginia Polytechnic State University
* Funded by Microsoft
* Date Updated: 7/29/2004
* AutoDataCollection is the application to automate testing process in order to measure
* performance of our wireless ad hoc messenger using the DSR protocol.
*/

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Xml;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.IO;
using System.Text;


namespace AutoDataCollection
{
	/// <summary>
	/// Summary description for AutoDataCollection.
	/// </summary>
	public class AutoDataCollection : System.Windows.Forms.Form
	{
		private ControlInvoker controlInvoker;
		private System.Windows.Forms.Label label1;
		private System.Windows.Forms.TextBox MyID_IP;
		private System.Windows.Forms.Button buddy_Btn;
		private System.Windows.Forms.CheckBox ACM_CB;
		private System.Windows.Forms.Label label3;
		private System.Windows.Forms.TextBox SWS_TB;
		private System.Windows.Forms.Button Start_Btn;
		private System.Windows.Forms.Label label4;
		private System.Windows.Forms.CheckBox Cache_CB;
		private System.Windows.Forms.Button Exit_Btn;
		private System.Windows.Forms.TextBox NumOfData_TB;
		private System.Windows.Forms.MainMenu mainMenu1;
		private System.Windows.Forms.TextBox BuddyID_IP;
		private System.Windows.Forms.Button Reset_Btn;
	
		//***********************************************************************//
		//UdpServer
		//***********************************************************************//
		private Socket udpSocket;
		private Thread udpListeningThread;
		public string ReceivedData;
		public int ReceivedDataLength;
	
		//********************************************************************//
		//UdpClient
		private UdpClient udpClient;
		private bool udpServerIsAlive;
		//***********************************************************************//
		//Variables to hold in User Interface
		public string strMyID="NA";
		public string strMyIP="NA";
		public string strBuddyID="NA";
		public string strBuddyIP="NA";
		private string strMessageBox="";		
		//Show buddyList in ListView
		private Hashtable hashBuddyList;
		//Route Cache to store identified paths for each destination node
		public RouteCache routeCache;
		
		//XmlDoc accessible from anywhere
		private XmlDocument xmlDoc;

		//One sequence for a certain ID
		//private ArrayList sequence;
		//It contains source ID as a Key and sequence ArrayList as a Value
		private Hashtable hashSequence;
		//ArrayList neighbors
		private ArrayList neighbors;
		//Request ID
		public int sequenceNum = 0;
		public Hashtable send_buffer;
		public Hashtable ReceiptBuffer;
		private Hashtable RRTimerBuffer;
		private Hashtable ReceiptTimerBuffer;
		private Hashtable source_neighbors;
		public Hashtable RRTimer2Buffer;
		public Hashtable ReceiptTimer2Buffer;
		public ArrayList OnCallingBuddy;
		
		private string MySessionID = ""; 
		private string BuddySessionID = "";
		public int processedPacketSN;
		private int discoverSeqNum;
		private int REPSeqNum;
		public DateTime startingDateTime;
		private bool OnACM;
		private bool OnCache;
		private bool isStartPacket;
		private int SWS;	
		private int NumOfDataPacket;					
		public int NumOfDataPacketSent = 0;
		//private double ElapsedTime = 0;				
		private ArrayList resultList;
		private string file_name ="";	
		//private bool isNotStarted = true;
		private ArrayList SenderWindow;
		private ArrayList ReceiverWindow;
		private bool isDoneConnectionEstablishment;
		private System.Windows.Forms.CheckBox RP_CB;
		public ArrayList processingRDBuffer;
		private bool OnReceiptPacket;
	
		public AutoDataCollection()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();

			try
			{
				//run UdpServer
				processingRDBuffer = new ArrayList();
				start_server();				
				controlInvoker = new ControlInvoker (this);	
				hashBuddyList = new Hashtable();
				xmlDoc = new XmlDocument();
				xmlDoc.Load(new StreamReader("\\My Documents\\Personal\\conf.xml"));
				hashSequence = new Hashtable();
				storeSequenceInHashtable();
				neighbors = new ArrayList();
				send_buffer = new Hashtable();
				ReceiptBuffer = new Hashtable(); 
				RRTimerBuffer = new Hashtable();
				RRTimer2Buffer = new Hashtable();
				ReceiptTimerBuffer = new Hashtable();
				ReceiptTimer2Buffer = new Hashtable();
				source_neighbors = new Hashtable();
				OnCallingBuddy = new ArrayList();
				processedPacketSN = -9;
				discoverSeqNum = -3;	
				REPSeqNum = 400;
				isStartPacket = true;
				resultList = new ArrayList();
				SenderWindow = new ArrayList();
				ReceiverWindow = new ArrayList();				
				isDoneConnectionEstablishment = false;
				OnReceiptPacket = false;
			}
			catch(Exception e)
			{
				MessageBox.Show("PocketPCChat() Exception is occurred: "+e.Message);
			}
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			base.Dispose( disposing );
		}

		static void Main(string[] args)
		{		
			Application.Run(new AutoDataCollection());
			//Application.Exit();
		}
	

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.label1 = new System.Windows.Forms.Label();
			this.MyID_IP = new System.Windows.Forms.TextBox();
			this.buddy_Btn = new System.Windows.Forms.Button();
			this.ACM_CB = new System.Windows.Forms.CheckBox();
			this.label3 = new System.Windows.Forms.Label();
			this.SWS_TB = new System.Windows.Forms.TextBox();
			this.Start_Btn = new System.Windows.Forms.Button();
			this.label4 = new System.Windows.Forms.Label();
			this.NumOfData_TB = new System.Windows.Forms.TextBox();
			this.Cache_CB = new System.Windows.Forms.CheckBox();
			this.Exit_Btn = new System.Windows.Forms.Button();
			this.BuddyID_IP = new System.Windows.Forms.TextBox();
			this.mainMenu1 = new System.Windows.Forms.MainMenu();
			this.Reset_Btn = new System.Windows.Forms.Button();
			this.RP_CB = new System.Windows.Forms.CheckBox();
			// 
			// label1
			// 
			this.label1.Location = new System.Drawing.Point(8, 8);
			this.label1.Size = new System.Drawing.Size(80, 20);
			this.label1.Text = "My ID and IP";
			// 
			// MyID_IP
			// 
			this.MyID_IP.Location = new System.Drawing.Point(96, 8);
			this.MyID_IP.ReadOnly = true;
			this.MyID_IP.Size = new System.Drawing.Size(136, 22);
			this.MyID_IP.Text = "";
			// 
			// buddy_Btn
			// 
			this.buddy_Btn.Location = new System.Drawing.Point(0, 40);
			this.buddy_Btn.Size = new System.Drawing.Size(88, 32);
			this.buddy_Btn.Text = "Select buddy";
			this.buddy_Btn.Click += new System.EventHandler(this.buddy_Btn_Click);
			// 
			// ACM_CB
			// 
			this.ACM_CB.Checked = true;
			this.ACM_CB.CheckState = System.Windows.Forms.CheckState.Checked;
			this.ACM_CB.Location = new System.Drawing.Point(8, 80);
			this.ACM_CB.Size = new System.Drawing.Size(224, 20);
			this.ACM_CB.Text = "All Connected with Multi-hops (ACM)";
			// 
			// label3
			// 
			this.label3.Location = new System.Drawing.Point(8, 168);
			this.label3.Size = new System.Drawing.Size(152, 16);
			this.label3.Text = "Sliding Window Size (SWS)";
			// 
			// SWS_TB
			// 
			this.SWS_TB.Location = new System.Drawing.Point(176, 168);
			this.SWS_TB.Size = new System.Drawing.Size(32, 22);
			this.SWS_TB.Text = "1";
			this.SWS_TB.TextChanged += new System.EventHandler(this.SWS_TB_TextChanged);
			// 
			// Start_Btn
			// 
			this.Start_Btn.Location = new System.Drawing.Point(16, 224);
			this.Start_Btn.Size = new System.Drawing.Size(64, 32);
			this.Start_Btn.Text = "Start";
			this.Start_Btn.Click += new System.EventHandler(this.Start_Btn_Click);
			// 
			// label4
			// 
			this.label4.Location = new System.Drawing.Point(8, 200);
			this.label4.Size = new System.Drawing.Size(152, 16);
			this.label4.Text = "Number of Data Packets";
			// 
			// NumOfData_TB
			// 
			this.NumOfData_TB.Location = new System.Drawing.Point(176, 200);
			this.NumOfData_TB.Size = new System.Drawing.Size(40, 22);
			this.NumOfData_TB.Text = "100";
			// 
			// Cache_CB
			// 
			this.Cache_CB.Checked = true;
			this.Cache_CB.CheckState = System.Windows.Forms.CheckState.Checked;
			this.Cache_CB.Location = new System.Drawing.Point(8, 104);
			this.Cache_CB.Size = new System.Drawing.Size(176, 24);
			this.Cache_CB.Text = "Use of Cache Structure";
			// 
			// Exit_Btn
			// 
			this.Exit_Btn.Location = new System.Drawing.Point(88, 224);
			this.Exit_Btn.Size = new System.Drawing.Size(64, 32);
			this.Exit_Btn.Text = "Exit";
			this.Exit_Btn.Click += new System.EventHandler(this.Exit_Btn_Click);
			// 
			// BuddyID_IP
			// 
			this.BuddyID_IP.Location = new System.Drawing.Point(96, 48);
			this.BuddyID_IP.ReadOnly = true;
			this.BuddyID_IP.Size = new System.Drawing.Size(136, 22);
			this.BuddyID_IP.Text = "";
			// 
			// Reset_Btn
			// 
			this.Reset_Btn.Location = new System.Drawing.Point(168, 224);
			this.Reset_Btn.Size = new System.Drawing.Size(64, 32);
			this.Reset_Btn.Text = "Reset";
			this.Reset_Btn.Click += new System.EventHandler(this.Reset_Btn_Click);
			// 
			// RP_CB
			// 
			this.RP_CB.Checked = true;
			this.RP_CB.CheckState = System.Windows.Forms.CheckState.Checked;
			this.RP_CB.Location = new System.Drawing.Point(8, 136);
			this.RP_CB.Size = new System.Drawing.Size(144, 24);
			this.RP_CB.Text = "Use of Receipt Packet";
			this.RP_CB.CheckStateChanged += new System.EventHandler(this.RP_CB_CheckStateChanged);
			// 
			// AutoDataCollection
			// 
			this.BackColor = System.Drawing.Color.YellowGreen;
			this.ClientSize = new System.Drawing.Size(240, 287);
			this.Controls.Add(this.RP_CB);
			this.Controls.Add(this.Reset_Btn);
			this.Controls.Add(this.BuddyID_IP);
			this.Controls.Add(this.Exit_Btn);
			this.Controls.Add(this.Cache_CB);
			this.Controls.Add(this.NumOfData_TB);
			this.Controls.Add(this.label4);
			this.Controls.Add(this.SWS_TB);
			this.Controls.Add(this.label3);
			this.Controls.Add(this.ACM_CB);
			this.Controls.Add(this.buddy_Btn);
			this.Controls.Add(this.MyID_IP);
			this.Controls.Add(this.label1);
			this.Controls.Add(this.Start_Btn);
			this.Menu = this.mainMenu1;
			this.Text = "AutoDataCollection";
			this.Load += new System.EventHandler(this.AutoDataCollection_Load);

		}
		#endregion

		/// <summary>
		/// When the frame is loaded, this method is implemented.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void AutoDataCollection_Load(object sender, System.EventArgs e)
		{
			try
			{
				//My IP Address
				IPHostEntry localHostEntry = Dns.GetHostByName(Dns.GetHostName());
				IPAddress myIP = localHostEntry.AddressList[0];
				strMyIP = myIP.ToString();				
				//My ID
				strMyID = FindIDbyIPAddressInConfFile(strMyIP);
				//Set MyID_IP
                MyID_IP.Text = strMyID+" : "+strMyIP;
				//Set routeCache
				routeCache = new RouteCache();
				//Set hashBuddyList
				setHashBuddyList();				
			}
			catch(Exception ex)
			{
				MessageBox.Show("AutoDataCollection_Load() Exception occurred : "+ex.Message);
			}

		}

		/// <summary>
		/// Set hashBuddyList(Hashtable), id as key and ip as value
		/// </summary>
		public void setHashBuddyList()
		{
			try
			{
				XmlNode root = xmlDoc.DocumentElement;
				XmlNodeList xNodes = root.ChildNodes;
				foreach(XmlNode childNode in xNodes)
				{					
					XmlElement currentElement = (XmlElement)childNode;
					string id = currentElement.GetAttribute("ID").ToString();
					string ip = currentElement.GetAttribute("IP").ToString();
					hashBuddyList.Add(id,ip);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in setHashBuddyList() : "+e.Message);
			}
		}

		/// <summary>
		/// Get Pocket PC ID by supplying its IP address
		/// </summary>
		/// <param name="strNodeIPAddress">Pocket PC IP Address in x.x.x.x format</param>
		/// <returns>ID matching input IP address</returns>
		public string FindIDbyIPAddressInConfFile(string strNodeIPAddress)
		{			
			string strReturnID="NA";
			try
			{
				XmlNode root = xmlDoc.DocumentElement;//configuration root element
				XmlNodeList xNodes = root.ChildNodes;
											
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement currentElement = (XmlElement)childNode;
					if(currentElement.GetAttribute("IP").Equals(strNodeIPAddress))
					{
						strReturnID=currentElement.GetAttribute("ID").ToString();
					}
				}									
			}
			catch(Exception ex)
			{
				MessageBox.Show("FindIDbyIPAddress: An Exception has occured. \r\n" + ex.ToString()+ex.Message.ToString(), 
					"Pocket Chat",MessageBoxButtons.OK , MessageBoxIcon.Hand ,MessageBoxDefaultButton.Button1);
			}
			return strReturnID;
		}

		/// <summary>
		/// Get IP address with ID input
		/// </summary>
		/// <param name="strID"></param>
		/// <returns>IP address matching with ID input</returns>
		public string GetIPAddressByIDInConfFile(string strID)
		{
			string strReturnIP="NA";
			try
			{
				XmlNode root = xmlDoc.DocumentElement;//configuration root element
				XmlNodeList xNodes = root.ChildNodes;
											
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement currentElement = (XmlElement)childNode;
					if(currentElement.GetAttribute("ID").Equals(strID))
					{
						strReturnIP=currentElement.GetAttribute("IP").ToString();
					}
				}									
			}
			catch(Exception ex)
			{
				MessageBox.Show("FindIDbyIPAddress: An Exception has occured. \r\n" + ex.ToString()+ex.Message.ToString(), 
					"Pocket Chat",MessageBoxButtons.OK , MessageBoxIcon.Hand ,MessageBoxDefaultButton.Button1);
			}
			return strReturnIP;
		}

		public ArrayList getNeighbors(string p_ID)
		{
			ArrayList neighbor = new ArrayList();
			Hashtable node_list = new Hashtable();
			
			try
			{
				NetworkStatus ns = new NetworkStatus(hashSequence);
				IDictionaryEnumerator ide = hashSequence.GetEnumerator();
				int T = getNTTime();
				while(ide.MoveNext())
				{
					string node_id = ide.Key.ToString();
					Point point = new Point();
					point = ns.getLocation(T, node_id);
					node_list.Add(node_id, point);
				}//while
				ns.adjustNetworkTopology(node_list, OnACM, isStartPacket);
				neighbor = ns.getNeighbors(strMyID);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getNeighbors() PocketPCChat : "+e.Message);
			}
			return neighbor;
		}

		private void Exit_Btn_Click(object sender, System.EventArgs e)
		{
			if(resultList.Count!=0)
			{
				writeResultXML();
				writeDataAnalysisXML();				
				InformToWriteResultXMLToAllNodes();
			}
			exit();			
		}//getNeightbors()

		/// <summary>
		/// When EXIT buttons are pressed, this method is called.
		/// </summary>
		private void exit()
		{
			udpServerIsAlive = false;
			stop_server();			
			resetAll();
			this.Close();
		}

		private void resetAll()
		{			
			SenderWindow.Clear();
			ReceiverWindow.Clear();
			resultList.Clear();			
			sequenceNum = 0;
			send_buffer.Clear();
			ReceiptBuffer.Clear();
			MySessionID = "";
			BuddySessionID = "";
			ReceiptTimerBuffer.Clear();
			processedPacketSN = -9;
			discoverSeqNum = -3;
			REPSeqNum = 400;
			OnCallingBuddy.Clear();
			ReceiptTimer2Buffer.Clear();
			RRTimer2Buffer.Clear();
			routeCache.Clear();	
			//ElapsedTime = 0;
		}

		/// <summary>
		/// Save the packet to send in send_buffer unless the send_buffer
		/// already has the packet and its request id or crrent node is not initiator.
		/// </summary>
		/// <param name="packet">packet passed</param>
		private void SaveInSendBuffer(Packet packet)
		{
			try
			{
				if(!send_buffer.ContainsKey(packet.strRequestID))
				{
					send_buffer.Add(packet.strRequestID, packet);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SaveInSendBuffer() :"+e.Message);
			}
		}

		private void SetIsStartPacket(Packet p_packet)
		{
			if(!p_packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET) &&
				!p_packet.strType.Equals(Packet_CONSTANTS.REJECT_START_CHAT_PACKET) &&
				!p_packet.strType.Equals(Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET))
			{
				isStartPacket = false;	
			}
		}
		
		/// <summary>
		/// In Send Button, Packet with a message is sent to target node(BuddyID)
		/// </summary>
		/// <param name="packet">Packet with a message</param>
		public void SendPacket(ref Packet packet)
		{
			try
			{			
				SetIsStartPacket(packet);				
				XmlDocument doc = new XmlDocument();
				doc.LoadXml("<Packet></Packet>");

				XmlNode root = doc.DocumentElement;

				//1. Type
				XmlElement elem = doc.CreateElement(Packet_CONSTANTS.TYPE);
				elem.InnerText = packet.strType;
				root.AppendChild(elem);


				//2.Initiator
				elem = doc.CreateElement(Packet_CONSTANTS.INITIATOR);
				elem.InnerText = packet.strInitiator;
				root.AppendChild(elem);


				//3.Current ID
				elem = doc.CreateElement(Packet_CONSTANTS.CURRENTID);
				elem.InnerText = packet.strCurrentID;
				root.AppendChild(elem);
				
				//4.Payload
				elem = doc.CreateElement(Packet_CONSTANTS.PAYLOAD);
				elem.InnerText = packet.strPayload;
				root.AppendChild(elem);

				//5. RequestID
				elem = doc.CreateElement(Packet_CONSTANTS.REQUESTID);
				elem.InnerText = packet.strRequestID;
				root.AppendChild(elem);

				//6.RouteCache
				elem = doc.CreateElement(Packet_CONSTANTS.ROUTECACHE);
				elem.InnerText = packet.strRouteCache;
				root.AppendChild(elem);

				//7.RouteRecord
				elem = doc.CreateElement(Packet_CONSTANTS.ROUTERECORD);
				elem.InnerText = packet.strRouteRecord;
				root.AppendChild(elem);

				//8.Target
				elem = doc.CreateElement(Packet_CONSTANTS.TARGET);
				elem.InnerText = packet.strTarget;
				root.AppendChild(elem);
				
				//9.ReceiptID
				elem = doc.CreateElement(Packet_CONSTANTS.RECEIPTID);
				elem.InnerText = packet.strReceiptID;
				root.AppendChild(elem);

				//for debugging
				string messageToSend = "";
				string serverID = "";
				string serverIP = "";
				bool sentSuccess = false;
				neighbors = getNeighbors(strMyID);					

				if(packet.strRouteCache.Equals(Packet_CONSTANTS.NOTAVAILABLE))
				{//when there is no available path in route cache, broadcast to 
					//all neighboring nodes

					if(neighbors.Count>0)
					{
						for(int i=0; i<neighbors.Count; i++)
						{
							serverID = neighbors[i].ToString();
							if(isProperNextNodeInRD(serverID, packet.strRouteRecord))
							{
								//9.Replace ReceiptID
								packet.strReceiptID = getReceiptID(strMyID, serverID);
								elem.InnerText = packet.strReceiptID;
								root.AppendChild(elem);

								//10. Put TimeStamp into packet
								elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
								if(packet.strInitiator.Equals(strMyID))
								{
									packet.strTimeStamp = getNTTime().ToString();
								}
								elem.InnerText = packet.strTimeStamp;
								root.AppendChild(elem);

								SaveInReceiptBuffer(packet);//for a receipt packet

								serverIP = GetIPAddressByIDInConfFile(serverID);
								messageToSend = doc.InnerXml.ToString();
//								showDebugSendingPackets(packet);
								ResultPacket rp = new ResultPacket();
								rp.strComType = Packet_CONSTANTS.SENT;
								storeResultPackets(rp, packet);
								sentSuccess = sendUDP(serverIP, messageToSend);
							}//Proper Next Node if
						}//end of for
					}
					else
						//neighbors.Count == 0: when there is no neighbors, 
						//send route error packet to the source
					{
						serverID = "NA";
						//9.Replace ReceiptID
						packet.strReceiptID = getReceiptID(strMyID, serverID);
						elem.InnerText = packet.strReceiptID;
						root.AppendChild(elem);

						//10. Put TimeStamp into packet
						elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
						if(packet.strInitiator.Equals(strMyID))
						{
							packet.strTimeStamp = getNTTime().ToString();
						}
						elem.InnerText = packet.strTimeStamp;
						root.AppendChild(elem);

						ResultPacket rp = new ResultPacket();
						rp.strComType = Packet_CONSTANTS.SENT;
						storeResultPackets(rp, packet);

						SaveInReceiptBuffer(packet);//for a receipt packet
						ReleaseReceiptTimer(packet.strReceiptID);
						InformRouteErrorToSource(packet);
					}
				}
				else
				{//When route cache has a path to reach target
									
					serverID = getNextNodeInCachePath(packet.strRouteCache);
					serverIP = GetIPAddressByIDInConfFile(serverID);

					//9.Replace ReceiptID
					packet.strReceiptID = getReceiptID(strMyID, serverID);
					elem.InnerText = packet.strReceiptID;
					root.AppendChild(elem);

					//10. Put TimeStamp into packet
					elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
					if(packet.strInitiator.Equals(strMyID))
					{
						packet.strTimeStamp = getNTTime().ToString();
					}
					elem.InnerText = packet.strTimeStamp;
					root.AppendChild(elem);

					ResultPacket rp = new ResultPacket();
					rp.strComType = Packet_CONSTANTS.SENT;
					storeResultPackets(rp, packet);

					SaveInReceiptBuffer(packet);//for a receipt packet
					if(neighbors.Contains(serverID))
					{
						messageToSend = doc.InnerXml.ToString();
						sentSuccess = sendUDP(serverIP, messageToSend);
					}
					else
					{
						ReleaseReceiptTimer(packet.strReceiptID);
						InformRouteErrorToSource(packet);
					}
				}
			}//try
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in SendPacket() : "+e.Message);
			}

		}//end of SendPacket()
	    
		private bool isProperNextNodeInRD(string p_serverID, string p_RouteRecord)
		{
			for(int i=0; i<p_RouteRecord.Length; i++)
			{
				if(p_serverID.Equals(p_RouteRecord[i].ToString()))
				{
					return false;
				}
			}
			return true;
		}

		private void SetMySessionID()
		{			
			MySessionID = getSessionID()+":"+strMyID+":"+strBuddyID;
		}

		/// <summary> 
		/// Check Receiver Buffer by using sequence number if duplicate packets are recieved.
		/// Return true if the request id is already received. Otherwise, return false.
		/// </summary>
		/// <param name="p_requestID">request ID of the received packet</param>
		/// <returns>boolean</returns>
		private bool isAlreadyReceived(Packet packet)
		{
			try
			{
				//for all kinds of route reply packets: In Sender
				if(packet.strType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.REJECT_START_CHAT_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET))
				{
					if(send_buffer.ContainsKey(packet.strRequestID))
					{
						return false;
					}
					return true; 
				}
				//packets from buddy side: In Receiver
				else if(packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET))
				{
					int receivedSN = getSequenceNumFromRequestIDReceived(packet.strRequestID);
					if(receivedSN<=processedPacketSN)
					{
						return true;
					}
					else
					{
                        return isInReceiverWindow(receivedSN.ToString());
					}					
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.ROUTE_ERROR_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
				{
						return false;
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
				{
					if(isDoneConnectionEstablishment)
					{
                       return true;
					}
					return false;
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
				{
					if(!isDoneConnectionEstablishment)
					{
                       return true;
					}
					return false;
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.ONCALLING_PACKET))
				{
					if(send_buffer.Contains(packet.strRequestID))
					{
						return false;
					}
					else
					{
						return true;
					}
				}
				return false;
			}//try			
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in isAlreadyReceived(): "+e.Message);
				return false;
			}
		}

		private void SetReceiptTimer(string p_receiptID, Packet p_packet)
		{
			try
			{
				MyTimer2 myTimer = new MyTimer2(1000, true,1,this,p_packet, p_receiptID);
				myTimer.SetReceiptTimer2();
				ReceiptTimer2Buffer.Add(p_receiptID, myTimer);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception i SetReceiptTimer(); "+e.Message);
			}
		}

		private void SaveInReceiptBuffer(Packet packet)
		{
			try
			{
				if(OnReceiptPacket)
				{	
					if(packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET))
					{				
						SetReceiptTimer(packet.strReceiptID, packet);
						ReceiptBuffer.Add(packet.strReceiptID, packet);
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SaveInReceiptBuffer(): "+e.Message);
			}
		}

		public void UpdateReceiptBuffer(string p_receiptID)
		{
			try
			{
				if(ReceiptBuffer.ContainsKey(p_receiptID))
				{
					ReceiptBuffer.Remove(p_receiptID);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in UpdateReceiptBuffer():"+e.Message);
			}
		}

		/// <summary>
		/// Release Receipt Timer2
		/// </summary>
		/// <param name="p_ID">receiptID of packet</param>
		public void ReleaseReceiptTimer(string p_ID)
		{
			try
			{
				IDictionaryEnumerator ide = ReceiptTimer2Buffer.GetEnumerator();
				while(ide.MoveNext())
				{
					if(ide.Key.ToString().Equals(p_ID))
					{
						MyTimer2 tempTimer = (MyTimer2)ide.Value;	
						tempTimer.setEnabled(false);
						break;
					}
				}
				ReceiptTimer2Buffer.Remove(p_ID);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception ocurred in ReleaseReceiptTimer():"+e.Message);
			}
		}
		private bool isReceiptPacket()
		{
			string startStringInReceiptPacket = "<Packet><Type>"+Packet_CONSTANTS.RECEIPT_PACKET;
			if(ReceivedDataLength>0)
			{
				if(ReceivedData.StartsWith(startStringInReceiptPacket))
				{
					return true;
				}
			}
			return false;
		}

		private BroadcastPacket transformXMLIntoBroadcastPacket()
		{
            BroadcastPacket rcv_packet = new BroadcastPacket();
			try
			{				
				XmlDocument doc = new XmlDocument();
				doc.LoadXml(ReceivedData);
				ReceivedDataLength=0;//reset the received data's length
				XmlNode root = doc.DocumentElement; 
				XmlNodeList xNodes = root.ChildNodes;
			
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals(Packet_CONSTANTS.TYPE))
					{
						rcv_packet.strType = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.CACHE))
					{
						rcv_packet.strOnCache = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.ACM))
					{
						rcv_packet.strOnACM = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.SWS))
					{
						rcv_packet.strSWS = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.NUMOFDATAPACKET))
					{
						rcv_packet.strNumOfDP = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.RP))
					{
						rcv_packet.strOnRP = elem.FirstChild.Value;
   					}
				}	
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in transformXmlIntoBroadcastPacket() : "+e.Message);
			}
			return rcv_packet;					
        }

		private ReceiptPacket transformXmlIntoReceiptPacket()
		{
			ReceiptPacket rcv_packet = new ReceiptPacket();
			try
			{
				XmlDocument doc = new XmlDocument();
				doc.LoadXml(ReceivedData);
				ReceivedDataLength=0;//reset the received data's length
				XmlNode root = doc.DocumentElement; 
				XmlNodeList xNodes = root.ChildNodes;
			
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals(Packet_CONSTANTS.TYPE))
					{
						rcv_packet.strType = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.RECEIPTID))
					{
						rcv_packet.strReceiptID = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.TIMESTAMP))
					{
						rcv_packet.strTimeStamp = elem.FirstChild.Value;
					}
				}	
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in transformXmlIntoReceiptPacket() : "+e.Message);
			}
			return rcv_packet;			
		}

		private void resetForHangUpPacket(Packet packet)
		{
			if(packet.strType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET))
			{
				resetAll();
			}
		}
		

		private bool isBroadcastNTPacket()
		{
			try
			{
				string startString = "<Packet><Type>"+Packet_CONSTANTS.BROADCAST_NT_PACKET;
			
				if(ReceivedData.StartsWith(startString))
				{
					return true; 
				}
				else
				{
					return false; 
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in isBroadcastNTPacket(): "+e.Message);
				return false;
			}
		}

		public int getNTTime()
		{
			long timeElapsed = (DateTime.Now.Ticks - startingDateTime.Ticks)/10000000;
			return int.Parse(timeElapsed.ToString());
		}

		private void startTickingNTTime()
		{
			startingDateTime = DateTime.Now;
		}

		private void SendOnCallingPacket(ref Packet packet)
		{
			try
			{
				string initiator = packet.strTarget;
				string target = packet.strInitiator;

				packet.strInitiator = initiator;
				packet.strTarget = target;

				if(!routeCache.isEmpty())
				{
					packet.strRouteCache = routeCache.getPath();
				}
				else
				{
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
				}
				packet.strRouteRecord = strMyID;
				packet.strCurrentID = strMyID;
				packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				packet.strType = Packet_CONSTANTS.ONCALLING_PACKET;
				SendPacket(ref packet);						    
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendOnCallingPacket(): "+e.Message);    
			}
		}
		


		private bool isProperBuddy(string p_initiator)
		{
			try
			{
				if(strBuddyID.Equals(p_initiator) || strBuddyID.Equals("NA"))
				{
					return true;
				}
				return false;
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in isProperBuddy():"+e.Message);
				return false;
			}
		}

		private void SetAllBasicInfo(BroadcastPacket broadPacket)
		{
			try
			{
                SWS_TB.Text = broadPacket.strSWS;
				if(broadPacket.strOnCache.Equals(true.ToString()))
				{
					Cache_CB.Checked = true;
				}
				else
				{
					Cache_CB.Checked = false;
				}
  
				if(broadPacket.strOnACM.Equals(true.ToString()))
				{
					ACM_CB.Checked = true;
				}
				else
				{
					ACM_CB.Checked = false;
				}

				if(broadPacket.strOnRP.Equals(true.ToString()))
				{
					RP_CB.Checked = true;
				}
				else
				{
					RP_CB.Checked = false;
				}
				
				NumOfData_TB.Text = broadPacket.strNumOfDP;

				SWS = int.Parse(SWS_TB.Text);
				OnCache = Cache_CB.Checked;
				OnACM = ACM_CB.Checked;			
				OnReceiptPacket = RP_CB.Checked;
				NumOfDataPacket = int.Parse(NumOfData_TB.Text);
				NumOfDataPacketSent = NumOfDataPacket;
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in SetAllBasicInfo(): "+e.Message);
			}
		}

		public void SendData()
		{
			try
			{
				if((NumOfDataPacketSent>0))
				{
					while((SenderWindow.Count<SWS) && (NumOfDataPacketSent>0))
					{
						strMessageBox = "DP"+NumOfDataPacketSent.ToString();
						NumOfDataPacketSent--;
						Packet packet = new Packet();
						neighbors = getNeighbors(strMyID);
						if(!routeCache.isEmpty())
						{
							SendMessagePacket(ref packet);
						}
						else if(neighbors.Count==0)
						{
                            NumOfDataPacketSent++;
						}
						else
						{
							packet.strType = Packet_CONSTANTS.DATA_PACKET;
							packet.strTarget = strBuddyID;
							packet.strInitiator = strMyID;
							packet.strRequestID = getRequestID(Packet_CONSTANTS.DATA_PACKET);
							packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
							packet.strRouteRecord = strMyID;
							packet.strPayload = strMessageBox;
							packet.strCurrentID = strMyID;
							SaveInSendBuffer(packet);
							string temp = Packet_CONSTANTS.NOTAVAILABLE;
							PerformRouteDiscovery(temp,packet);
						}
					}
				}
				else//All data packets are sent. 
				{
				  //inform to write result.xml file to all nodes
					if(SenderWindow.Count == 0)
					{
						InformToWriteResultXMLToAllNodes();
					}
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show("Exception in SendData():"+ex.Message);
				MessageBox.Show("SenderWindow.Count: "+SenderWindow.Count);
			}			
		}

		public void InformToWriteResultXMLToAllNodes()
		{
			try
			{
                string strWriteResult = Packet_CONSTANTS.INFORM_WRITING_RESULTXML;
				IDictionaryEnumerator ide = hashBuddyList.GetEnumerator();
				while(ide.MoveNext())
				{
					string serverID = ide.Key.ToString();
					string serverIP = GetIPAddressByIDInConfFile(serverID);
					sendUDP(serverIP, strWriteResult);
				}	
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in InformToWriteResultXMLToAllNodes(): "+e.Message);
			}
		}

		private bool isInformingToWriteResultXML()
		{
			try
			{
				string startString = Packet_CONSTANTS.INFORM_WRITING_RESULTXML;
				if(ReceivedData.StartsWith(startString))
				{
					return true; 
				}
				else
				{
					return false; 
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in isInformingToWriteResultXML(): "+e.Message);
				return false;
			}
		}

		/// <summary>
		/// Generate result.xml by using resultList (ArrayList) containing 
		/// ResultPacket objects
		/// </summary>
		private void writeResultXML()
		{
			try
			{
				file_name = "\\My Documents\\Personal\\Result_"+strMyID+".xml";
                XmlTextWriter resultWriter = new XmlTextWriter(file_name,Encoding.ASCII);
					                             
				resultWriter.Formatting = Formatting.Indented;
				resultWriter.WriteStartDocument();
												
				resultWriter.WriteStartElement("Result");//root <Result>
				resultWriter.WriteAttributeString("NodeID", strMyID);
				resultWriter.WriteAttributeString("NodeIP", strMyIP);
				resultWriter.WriteAttributeString("TimeNow", getNTTime().ToString());
				for(int i = 0; i<resultList.Count; i++)
				{
					ResultPacket rp = (ResultPacket)resultList[i];
					resultWriter.WriteStartElement("Packet");//<Packet>
					resultWriter.WriteAttributeString("ComType",rp.strComType);
					resultWriter.WriteElementString("Type",rp.strType);
					resultWriter.WriteElementString("RequestID",rp.strRequestID);
					resultWriter.WriteElementString("RouteCacheUsed",rp.strRouteCacheUsed);
					resultWriter.WriteElementString("NumOfRoutesInCache",rp.strNumOfRoutesInCache);
					resultWriter.WriteElementString("TimeStampInSender",rp.strTimeStampInSender);                    
					resultWriter.WriteElementString("TimeStampInInReceiver",rp.strTimeStampInReceiver);                    
                    resultWriter.WriteElementString("NotTransmittedDP",rp.strNotTransmittedDP);
					resultWriter.WriteEndElement();//</Packet>                   
				}				
				resultWriter.WriteEndElement();//</Result>
				resultWriter.WriteEndDocument();
				resultWriter.Close();
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in writeResultXML(): "+e.Message);
			}
		}

		private void storeResultPackets(ResultPacket p_rp, Packet p_packet)
		{
			try
			{
				if(p_packet.strInitiator.Equals(strMyID) || 
					p_packet.strTarget.Equals(strMyID))
				{
					if(p_packet.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET) ||
						p_packet.strType.Equals(Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET))
					{					
						p_rp.strNotTransmittedDP =  p_packet.strPayload;
					}
					else if(p_packet.strType.Equals("REP"))
					{
						p_rp.strNotTransmittedDP = p_packet.strRequestID;
					}
					else
					{
						p_rp.strNotTransmittedDP = Packet_CONSTANTS.NOTAVAILABLE;
					}

					if(!(p_packet.strType.Equals("REP") 
						&& p_rp.strComType.Equals(Packet_CONSTANTS.RECEIVED)))
					{
						p_rp.strRequestID = p_packet.strRequestID;
					}

					p_rp.strRouteCacheUsed = p_packet.strRouteCache;
					p_rp.strTimeStampInSender = p_packet.strTimeStamp;
					if(p_rp.strComType.Equals(Packet_CONSTANTS.SENT))
					{
						p_rp.strTimeStampInReceiver = Packet_CONSTANTS.NOTAVAILABLE; 
					}
					else
					{
						p_rp.strTimeStampInReceiver = getNTTime().ToString();
					}
					p_rp.strType = p_packet.strType;
					p_rp.strNumOfRoutesInCache = routeCache.getNumOfRoutesInCache();
					resultList.Add(p_rp);
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in storeResultPacket(): "+e.Message);
			}
		}

		private void storeReceiptPackets(ResultPacket p_rp, ReceiptPacket p_receiptPacket)
		{
			try
			{                
				p_rp.strNumOfRoutesInCache = routeCache.getNumOfRoutesInCache();
				p_rp.strRequestID = p_receiptPacket.strReceiptID;
				p_rp.strRouteCacheUsed = Packet_CONSTANTS.NOTAVAILABLE;
				p_rp.strType = p_receiptPacket.strType;
				p_rp.strTimeStampInSender = p_receiptPacket.strTimeStamp;
				p_rp.strTimeStampInReceiver = getNTTime().ToString();
				resultList.Add(p_rp);
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in storeReceiptPackets():"+e.Message);
			}
		}

		private void SetAllButtonsDisabled()
		{
			Start_Btn.Enabled = false;
			buddy_Btn.Enabled = false;
			Reset_Btn.Enabled = false;
			ACM_CB.Enabled = false;
			Cache_CB.Enabled = false;
			RP_CB.Enabled = false;
			SWS_TB.ReadOnly = true;
			NumOfData_TB.ReadOnly = true;
		}

		private void SetAllInputsBackToBeginningState()
		{
			Start_Btn.Enabled = true;
			buddy_Btn.Enabled = true;
			Reset_Btn.Enabled = true;											
			ACM_CB.Enabled = true;
			Cache_CB.Enabled = true;
			RP_CB.Enabled = true;
			SWS_TB.ReadOnly = false;
			NumOfData_TB.ReadOnly = false;
			BuddyID_IP.Text = "";			
			ACM_CB.Checked = true;
			Cache_CB.Checked = true;
			SWS_TB.Text ="1";
			NumOfData_TB.Text = "100";
		}
		/// <summary>
		/// According to the Packet, each received packet is handled (receive/send) 
		/// appropriately.
		/// </summary>
		//private void HandleReceivedPacket()
		private void HandleReceivedPacket(object[] arguments)
		{
			try
			{				
				if(ReceivedDataLength>0)
				{
					if(isReceiptPacket())
					{						
						ReceiptPacket receivedPacket = transformXmlIntoReceiptPacket();	
						ReleaseReceiptTimer(receivedPacket.strReceiptID);
						UpdateReceiptBuffer(receivedPacket.strReceiptID);		
					}
					else if(isBroadcastNTPacket())
					{
						BroadcastPacket broadPacket = transformXMLIntoBroadcastPacket();
						startTickingNTTime();
						resetAll();			
						SetAllBasicInfo(broadPacket);							
						
						if(!BuddyID_IP.Text.Equals(""))//in source node
						{							
							writeBasicInfoXML();	
							Packet packet = new Packet();
							SetMySessionID();
							processedPacketSN = 0;
							SendStartChatPacket(ref packet); 
						}
						SetAllButtonsDisabled();						
					}
					else if(isInformingToWriteResultXML())
					{
						writeResultXML();
						writeDataAnalysisXML();
						SetAllInputsBackToBeginningState();
						exit();
					}
					else//All other packets
					{
						//SendReceiptPacket to previous node
						Packet receivedPacket = transformXmlIntoPacket();
						string previousNodeID = receivedPacket.strCurrentID;
						if(OnReceiptPacket)
						{
							//MessageBox.Show("Send RP: "+OnReceiptPacket.ToString());
							SendReceiptPacket(receivedPacket, previousNodeID);
						}
							
						if(strMyID.Equals(receivedPacket.strTarget))//when current node is the target node
						{							
							//MessageBox.Show("ReceivedData: \r\n"+ReceivedData);
							//if the packet is not received yet, store it. if the packet is already received but
							//the path is new, then store it. Otherwise, do nothing.
							receivedPacket.strRouteRecord = receivedPacket.strRouteRecord+":"+strMyID;
							StoreInCacheOfTargetNode(receivedPacket);//store in cache if necessary
					
							if(!isAlreadyReceived(receivedPacket))
							{
								ResultPacket rp = new ResultPacket();
								rp.strComType = Packet_CONSTANTS.RECEIVED;
								string REPRID ="";
								if(receivedPacket.strType.Equals("REP"))
								{
                                   rp.strRequestID = getRequestID(receivedPacket.strType);
								   REPRID = rp.strRequestID;
								}
								storeResultPackets(rp, receivedPacket);

								//1. Received "Start Chat Message"
								if(receivedPacket.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
								{

									if(isProperBuddy(receivedPacket.strInitiator))//(isProperBuddy(receivedPacket.strRequestID))
									{
										SetBuddySessionID(receivedPacket);
										isDoneConnectionEstablishment = true;               
										strBuddyID = receivedPacket.strInitiator;
										strBuddyIP = GetIPAddressByIDInConfFile(strBuddyID);										
										SetMySessionID();
										SendReplyStartChatPacket(ref receivedPacket);				
										BuddyID_IP.Text = strBuddyID+" : "+strBuddyIP;
									}
									else//not proper buddy
									{
										OnCallingBuddy.Add(receivedPacket.strRequestID);
										SendOnCallingPacket(ref receivedPacket);                       
									}
								}
								    //2. Received "Hang Up Message"   
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
								{
									receivedPacket.strRouteCache = getReverseRoute(receivedPacket.strRouteCache);
									SendRouteReplyPacket(ref receivedPacket);
									isDoneConnectionEstablishment = false;
									resetAll();
								}
							        //3. Received "Data Message"							
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.DATA_PACKET))									
								{
									AddToReceiverWindow(receivedPacket);
									SetBuddySessionID(receivedPacket);
									receivedPacket.strRouteCache = getReverseRoute(receivedPacket.strRouteCache);
									SendRouteReplyPacket(ref receivedPacket);
									ProcessPacketsInReceiverWindow();
								}
									//4.  Received "Reject Start Chat Message"  as Route Reply Packet
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.REJECT_START_CHAT_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									string initiator = receivedPacket.strInitiator;	
									MessageBox.Show(initiator+" does not want to talk with you");
									resetAll();
								}
									//5.  Received "Accept Start Chat Message"  as Route Reply Packet
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET))
								{			
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									string initiator = receivedPacket.strInitiator;																			
									SendData(); 
								}
									//6. Received "Route Reply Message" 
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET))									
								{
									UpdateSenderWindow(receivedPacket.strRequestID);
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									SendData();
								}							
									//7. "Reply Hang Up Message"
								if(receivedPacket.strType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);								
									resetForHangUpPacket(receivedPacket);
								}
									//8. Received "Route Error Message" 
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ROUTE_ERROR_PACKET))
								{
									//delete the path which had link failure
									ReleaseRRTimer(receivedPacket.strRequestID);
									
									if(send_buffer.Contains(receivedPacket.strRequestID))
									{
										Packet packet = getLostPacketFromSendBuffer(receivedPacket.strRequestID);
										if(routeCache.hasPath(packet.strRouteCache) )
										{
											routeCache.RemovePath(packet.strRouteCache);
										}				
										RepairRouteError(REPRID, ref receivedPacket);
									}
								}
									//9. Received "Route Discovery Packet"
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
								{
									SetBuddySessionID(receivedPacket);
									receivedPacket.strRouteCache = getReverseRoute(receivedPacket.strRouteRecord);
									SendRouteReplyPacket(ref receivedPacket);
								}
									//10. Received "Reply Route Discovery Packet"
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									processingRDBuffer.Remove(receivedPacket.strPayload);
									string lostPacketRequestID =receivedPacket.strPayload;
									if(send_buffer.Contains(lostPacketRequestID) && 
										!RRTimer2Buffer.Contains(lostPacketRequestID))
									{										
										Packet packetInSB = getLostPacketFromSendBuffer(lostPacketRequestID);
										string timerID = packetInSB.strRequestID;
										//Packet packetToResend = createPacketForRetransmission(ref packetInSB);
										packetInSB.strRouteCache = getReverseRoute(receivedPacket.strRouteCache); 
										UpdateSendBuffer(packetInSB.strRequestID);//remove old DP
										send_buffer.Add(packetInSB.strRequestID, packetInSB);//add new DP with new routeCache
										if(!RRTimer2Buffer.Contains(timerID))
										{
											SetRouteReplyTimer(timerID, packetInSB);									
											SendPacket(ref packetInSB);           
										}
									}
								}
									//11. Received "OnCalling Packet"
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ONCALLING_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									resetAll();
									MessageBox.Show(receivedPacket.strInitiator+" is busy!.");										
								}
								//}
							}// not aready received						
							else//the packet is already received.: For debugging purpose
							{
								if(receivedPacket.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
								{
									SendReplyStartChatPacket(ref receivedPacket);
								}
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
										receivedPacket.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
								{
									receivedPacket.strRouteCache = getReverseRoute(receivedPacket.strRouteRecord);
									SendRouteReplyPacket(ref receivedPacket);
								}
							}//already received packet
			
						}//if current node is target
						else//when current node is not the target node
						{
							if(receivedPacket.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
							{
								resetAll();
							}
							ForwardPacketToNextNode(ref receivedPacket);
						}

					}//not receipt packet						
				}//received data > 0
			}//try
			catch(Exception e)
			{
				MessageBox.Show("Exception is HandleReceivedPacket(): "+e.Message);
				MessageBox.Show("ReceivedData: "+ReceivedData);
			}
		}
		
		private Packet createPacketForRetransmission(ref Packet p_packet)
		{
			Packet packet = new Packet();
			try
			{
				packet.strInitiator = p_packet.strInitiator;
				packet.strTarget = p_packet.strTarget;
				packet.strRequestID = p_packet.strRequestID;
				packet.strCurrentID = strMyID;
				packet.strPayload = p_packet.strPayload;
				packet.strRouteRecord = strMyID;
				packet.strType = p_packet.strType;
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in createPacketForRetransmission(): "+e.Message);
			}
			return packet;
		}

		private void SetBuddySessionID(Packet p_packet)
		{
			if(BuddySessionID.Equals(""))
			{
				strBuddyID = p_packet.strInitiator;
				BuddySessionID = getSessionIDFromRequestID(p_packet.strRequestID); 
			}
		}

		public string getPayloadFromReceiverWindow(string p_SN)
		{
			string payload = "N/A";
			try
			{
                IEnumerator ie = ReceiverWindow.GetEnumerator();
				while(ie.MoveNext())
				{
                    ElementsForReceiverWindow ERW = (ElementsForReceiverWindow)ie.Current;
					if(ERW.SN.Equals(p_SN))
					{
                        payload = ERW.Payload;
						break;
					}
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in getPacketFromReceiverWindow(): "+e.Message);
			}
			return payload;
		}
		/// <summary>
		/// Show the data in order from ReceiverWindow
		/// </summary>
		public void ProcessPacketsInReceiverWindow()
		{
			try
			{
								
				while(ReceiverWindow.Count>0) 
				{
					int nextPacketSN = processedPacketSN+1;
					if(isInReceiverWindow(nextPacketSN.ToString()))
					{
						ElementsForReceiverWindow ERW = 
							(ElementsForReceiverWindow)ReceiverWindow[0];
						if(ERW.SN.Equals(nextPacketSN.ToString()))
						{
							processedPacketSN = nextPacketSN;
							//In prototype, you need to display the message
							ReceiverWindow.RemoveAt(0);
						}
						else
						{
                            MessageBox.Show("Hello something wrong in ProcessPacketsInReceiverWindow()");
						}
					}
					else
					{
						break;
                        //Need to wait until the nextPacketSN is available 
						//by seeing the next arriving packet
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in ProcessPacketsInReceiverWindow() : "+e.Message);
			}
		}

		public int getSequenceNumFromRequestIDReceived(string p_requestID)
		{
			string requestID = p_requestID;
			string SID="";
			try
			{
				SID = getSessionIDFromRequestID(p_requestID);
				string strSeq = p_requestID.Substring(SID.Length+1);
				int new_seq = int.Parse(strSeq);
				return new_seq;
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getSequenceNumFromRequestIDReceived(): "+e.Message+" requestID is "+requestID+" SID is "+SID);
				return -100;
			}
		}


		/// <summary>
		/// Receiver node needs to get the SessionID which is determined by the sender node.
		/// </summary>
		/// <param name="p_requestID"></param>
		/// <returns></returns>
		private string getSessionIDFromRequestID(string p_requestID)
		{
			string SID = "";
			try
			{
				char colon = ':';
				int numOfColon = 0;
				for(int i = 0; i<p_requestID.Length; i++)
				{
					if(p_requestID[i].Equals(colon))
					{
						numOfColon++;
					}
					if(numOfColon <3)
					{
						SID = SID+p_requestID[i].ToString();
					}
					else
					{
						break;
					}

				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getSessionIDFromRequestID(): "+e.Message);
			}

			return SID;
		}

		/// <summary>
		/// Send a receipt packet to the previous node unless the packet's type is
		/// not a receipt packet.
		/// </summary>
		/// <param name="packet">packet</param>
		/// <param name="p_previousNodeID">previous node</param>
		private void SendReceiptPacket(Packet packet, string p_previousNodeID)
		{
			try
			{
				if(packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET))
				{

					ReceiptPacket rctPacket = new ReceiptPacket();
					rctPacket.strType = Packet_CONSTANTS.RECEIPT_PACKET;
					rctPacket.strReceiptID = packet.strReceiptID;
					rctPacket.strTimeStamp = getNTTime().ToString();
                
					XmlDocument doc = new XmlDocument();
					doc.LoadXml("<Packet></Packet>");

					XmlNode root = doc.DocumentElement;

					//1. Type
					XmlElement elem = doc.CreateElement(Packet_CONSTANTS.TYPE);
					elem.InnerText = rctPacket.strType;
					root.AppendChild(elem);

					//2.ReceiptID
					elem = doc.CreateElement(Packet_CONSTANTS.RECEIPTID);
					elem.InnerText = rctPacket.strReceiptID;
					root.AppendChild(elem);

					//3. TimeStamp
					elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
					elem.InnerText = rctPacket.strTimeStamp;
					root.AppendChild(elem);

					string messageToSend = doc.InnerXml.ToString();
					string serverID = p_previousNodeID;
					neighbors = getNeighbors(strMyID);					
					if(neighbors.Contains(serverID))
					{
						string serverIP = GetIPAddressByIDInConfFile(serverID); 
						bool sentSuccess = sendUDP(serverIP, messageToSend);
					}
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show("Exception is occurred in SendReceiptPacket(): "+ex.Message);
			}
		}

		
		private void SendRouteReplyPacket(ref Packet packet)
		{
			try
			{
				if(packet.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
				{
					packet.strType = Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET;				
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
				{
					packet.strType = Packet_CONSTANTS.REPLY_HANG_UP_PACKET;
				}
				else
				{
					packet.strType = Packet_CONSTANTS.ROUTE_REPLY_PACKET;
				}

				string target = packet.strInitiator;
			
				if(packet.strRouteCache.Equals(Packet_CONSTANTS.NOTAVAILABLE))
				{
					if(!routeCache.isEmpty())
					{
						packet.strRouteCache = routeCache.getPath();
					}
				}
				packet.strInitiator = strMyID;
				packet.strTarget = target;
				packet.strCurrentID = strMyID;
				if(!packet.strType.Equals(Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET))
				{
					packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				}
				//packet's request ID remains same. it will be checked in the original initiator
				packet.strRouteRecord = strMyID;
				SendPacket(ref packet);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in SendRouteReplyPacket(): "+e.Message);
			}
		}

		private string getReceiptID(string p_currentID, string p_nextID)
		{            
			string receiptID = "";
			Random receiptNum = new Random();
			int id = -1;
			while(true)
			{
				id = receiptNum.Next(100000);
				receiptID = p_currentID+":"+p_nextID+":"+id.ToString();
				if(!ReceiptBuffer.ContainsKey(receiptID))
				{
					break;  
				}
			}
			return receiptID;
		}

		public void SendStartChatPacket(ref Packet packet)
		{
			try
			{
				packet.strCurrentID = strMyID;
				packet.strInitiator = strMyID;
				packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				packet.strTarget = strBuddyID;
				packet.strType = Packet_CONSTANTS.START_CHAT_PACKET;
				packet.strRouteRecord = strMyID;
				if(!routeCache.isEmpty())
				{
					packet.strRouteCache = routeCache.getPath();
				}
				else
				{
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
				}		    
				packet.strRequestID = getRequestID(packet.strType);
				SaveInSendBuffer(packet);
				string timerID = packet.strRequestID;
				if(!RRTimer2Buffer.Contains(timerID))
				{
					SetRouteReplyTimer(timerID, packet);
					SendPacket(ref packet);				
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendStartChatPacket() : "+e.Message);
			}
		}

		/// <summary>
		/// Modified
		/// </summary>
		/// <param name="packet"></param>
		public void SendMessagePacket(ref Packet packet)
		{
			try
			{				
				packet.strInitiator = strMyID;
				packet.strCurrentID = strMyID;
				packet.strPayload = strMessageBox;
				packet.strTarget = strBuddyID;
				if(!routeCache.isEmpty())
				{
					packet.strRouteCache = routeCache.getPath();	
				}
				else
				{
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;			
				}
				packet.strRouteRecord = strMyID;
				packet.strType = Packet_CONSTANTS.DATA_PACKET;
				packet.strRequestID = getRequestID(packet.strType);
								
				if(SenderWindow.Count<SWS)
				{
					SenderWindow.Add(packet.strRequestID);
					SaveInSendBuffer(packet);
					string timerID = packet.strRequestID;
					/////MODIFIED////////////////////////
					if(!RRTimer2Buffer.Contains(timerID))
					{
						SetRouteReplyTimer(timerID, packet);
						SendPacket(ref packet);	
					}
					//////////////////////////////////////
				}
				else
				{
					MessageBox.Show("SendMessagePacket(): Please slow down to send messages.");
					MessageBox.Show(SenderWindow.Count+" "+showSenderWindow());
					//Allowed sender window size is SWS.  When user is sending more than SWS
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendMessagePacket():"+e.Message);
			}
		}

		private string showSenderWindow()
		{
			string inside = "N/A";
			for(int i =0; i<SenderWindow.Count; i++)
			{
				if(i == 0)
				{
                   inside = SenderWindow[i].ToString()+" ";
				}
				inside += SenderWindow[i].ToString()+" ";
			}
			return inside;
		}

		private Packet getPacketInSendBuffer()
		{
			Packet packet = new Packet();
			IEnumerator ie = send_buffer.Keys.GetEnumerator();
			string smallestRID = "";
			while(ie.MoveNext())
			{
				string comp = ie.Current.ToString();
				if(smallestRID.Equals(""))
				{
					smallestRID = comp;
				}
				else
				{
					if(smallestRID.CompareTo(comp)>0)
					{
						smallestRID = comp;
					}
				}				           
			}
			IDictionaryEnumerator ide = send_buffer.GetEnumerator();
			while(ide.MoveNext())
			{
				if(ide.Key.ToString().Equals(smallestRID))
				{
					packet = (Packet)ide.Value;
					break;
				}
			}
			return packet;
		}

		private void SendHangUpPacket(ref Packet packet)
		{
			try
			{
				packet.strType = Packet_CONSTANTS.HANG_UP_PACKET;
				packet.strCurrentID = strMyID;
				packet.strInitiator = strMyID;
				packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				packet.strTarget = strBuddyID;
				packet.strRouteRecord = strMyID;
				if(!routeCache.isEmpty())
				{
					packet.strRouteCache = routeCache.getPath();
				}
				else
				{
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
				}		
				packet.strRequestID = getRequestID(packet.strType);				
				
				while(true)
				{
					if(SenderWindow.Count==0)
					{
						SaveInSendBuffer(packet);
						string timerID = packet.strRequestID;
						if(!RRTimer2Buffer.Contains(timerID))
						{
							SetRouteReplyTimer(timerID, packet);
							SendPacket(ref packet);	
						}
						break;
					}
					else
					{
						Thread.Sleep(1000);
					}			
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendHangUpPacket():"+e.Message);
			}
		}

		/// <summary>
		/// With new Timer2
		/// </summary>
		/// <param name="p_ID"></param>
		/// <param name="p_packet"></param>
		private void SetRouteReplyTimer(string p_ID, Packet p_packet)
		{
			//for connection establishment packet, allow 5 seconds to get 
			//the answer for the chat request

			try
			{
				int time_interval = 5000;
				MyTimer2 myTimer = new MyTimer2(time_interval,true,3,this,p_packet,p_ID);
				myTimer.SetRRTimer2();
				RRTimer2Buffer.Add(p_ID, myTimer);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SetRouteReplyTimer(): "+e.Message+" id is "+p_ID+" packet's type is "+p_packet.strType);
			}            
		}


		private void SendReplyStartChatPacket(ref Packet packet)
		{
			try
			{
				string initiator = packet.strTarget;
				string target = packet.strInitiator;

				packet.strInitiator = initiator;
				packet.strTarget = target;
				packet.strRouteCache = getReverseRoute(packet.strRouteRecord);
				packet.strRouteRecord = strMyID;
				packet.strCurrentID = strMyID;
				packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
   
				packet.strType = Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET;  
				SendPacket(ref packet);						
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendReplyStartChatPacket(): "+e.Message);
			}
		}
		
	
		/// <summary>Debug is done
		/// If the received packet is a reply packet by checking an original request ID 
		/// in send buffer, the received reply packet is removed in the send buffer.
		/// </summary>
		/// <param name="p_RequestID"></param>
		public void UpdateSendBuffer(string p_RequestID)
		{
			try
			{
				if(send_buffer.Contains(p_RequestID))
				{
					send_buffer.Remove(p_RequestID);
				}				
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception ocurred in UodateSendBuffer() : "+e.Message);
			}
		}

		public void ReleaseRRTimer(string p_RequestID)
		{
			try
			{
				IDictionaryEnumerator ide = RRTimer2Buffer.GetEnumerator();
				while(ide.MoveNext())
				{
					if(ide.Key.ToString().Equals(p_RequestID))
					{
						MyTimer2 tempTimer = (MyTimer2)ide.Value;
						tempTimer.setEnabled(false);
						break;
					}
				}
				RRTimer2Buffer.Remove(p_RequestID);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in ReleaseRRTimer(): "+e.Message);
			}
		}

		/// <summary>
		/// When the current node is a target node, route cache is set appropriately
		/// : route destination and path in cache of current node
		/// </summary>
		/// <param name="packet">received packet</param>
		private void StoreInCacheOfTargetNode(Packet packet)
		{
			try
			{
				if(packet.strTarget.Equals(strMyID))
				{
					if(isProperBuddy(packet.strInitiator))
					{
						if(!(packet.strType.Equals(Packet_CONSTANTS.ROUTE_ERROR_PACKET)))
						{
							string reverseRoute  = getReverseRoute(packet.strRouteRecord);
							routeCache.AddPath(reverseRoute, OnCache);
						}
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in StoreInCacheOfTargetNode() : "+e.Message);
			}
		}

		/// <summary>
		/// Return a reverse route for entered route. Debug is done!
		/// </summary>
		/// <param name="route">Route</param>
		/// <returns></returns>
		private string getReverseRoute(string route)
		{
			char[] temp = route.ToCharArray();
			char[] reverseRoute = new char[temp.Length];
			int length = temp.Length-1;
			for(int i=0; i<reverseRoute.Length; i++)
			{
				reverseRoute[i] = temp[length-i];
			}
			string strRvsRoute = "";
			for(int i=0; i<reverseRoute.Length; i++)
			{
				strRvsRoute = strRvsRoute+reverseRoute[i].ToString();	
			}                
			return strRvsRoute;
		}

		/// <summary>
		/// In forwarding a packet, get the previous node ID.
		/// </summary>
		/// <param name="p_RouteRecord"></param>
		/// <returns>string which is a previous node ID</returns>
		private string getPreviousNode(string p_RouteRecord)
		{
			for(int i=2; i<p_RouteRecord.Length; i++)
			{
				if(strMyID.Equals(p_RouteRecord[i].ToString()))
				{
					return p_RouteRecord[i-2].ToString(); 
				}
			}
			return null;
		}

		private void ForwardPacketToNextNode(ref Packet packet)
		{
			packet.strRouteRecord = packet.strRouteRecord+":"+strMyID;
			packet.strCurrentID = strMyID;
			SendPacket(ref packet);
		}


		/// <summary>
		/// After receiving the packet, transform the Xml file to Packet object.
		/// </summary>
		/// <returns>Packet Object containing all information of xml data</returns>
		private Packet transformXmlIntoPacket()
		{
			Packet rcv_packet = new Packet();
			try
			{
				XmlDocument doc = new XmlDocument();
				doc.LoadXml(ReceivedData);
				ReceivedDataLength=0;//reset the received data's length
				XmlNode root = doc.DocumentElement; 
				XmlNodeList xNodes = root.ChildNodes;
			
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals(Packet_CONSTANTS.TYPE))
					{
						rcv_packet.strType = elem.FirstChild.Value;						
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.TARGET))
					{
						rcv_packet.strTarget = elem.FirstChild.Value;
					}
					else if (elem.Name.Equals(Packet_CONSTANTS.PAYLOAD))
					{
						rcv_packet.strPayload = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.CURRENTID))
					{
						rcv_packet.strCurrentID = elem.FirstChild.Value; 

					}
					else if(elem.Name.Equals(Packet_CONSTANTS.INITIATOR))
					{
						rcv_packet.strInitiator = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.ROUTECACHE))
					{
						rcv_packet.strRouteCache = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.REQUESTID))
					{
						rcv_packet.strRequestID = elem.FirstChild.Value;

					}
					else if(elem.Name.Equals(Packet_CONSTANTS.ROUTERECORD))
					{
						rcv_packet.strRouteRecord = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.RECEIPTID))
					{
						rcv_packet.strReceiptID = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.TIMESTAMP))
					{
						rcv_packet.strTimeStamp = elem.FirstChild.Value;
					}
				}	
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in transformXmlIntoPacket() : "+e.Message);
			}
			return rcv_packet;			
		}

		/// <summary>
		/// Get the next node in the strRouteCache
		/// </summary>
		/// <param name="p_path">appropriate path is passed</param>
		/// <returns>the next node to go</returns>
		private string getNextNodeInCachePath(string path)
		{

			string currentNode = strMyID;
			try
			{
				int index = path.IndexOf(currentNode);
				int nextNodeStartIndex = index+currentNode.Length+1;
				string subStr = path.Substring(nextNodeStartIndex);
			
				int colon = subStr.IndexOf(":");
				if(colon>0)
				{
					return subStr.Substring(0, colon);
				}
				else
				{
					return subStr;
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getNextNodeInCachePath():"+e.Message+" path: "+path); 
				return null;
			}
		}
		/// <summary>
		/// In Hashtable hashSequence, source as Key and sequence as Value are stored.
		/// </summary>
		public void storeSequenceInHashtable()
		{
			try
			{
				XmlNode root = xmlDoc.DocumentElement;//configuration root element
				XmlNodeList xNodes = root.ChildNodes;
											
				foreach(XmlNode childNode in xNodes)//<Node>
				{
					XmlElement currentElement = (XmlElement)childNode;
					//add to the hashSequence
					string source = currentElement.GetAttribute("ID").ToString();
					XmlNodeList SequenceNodes = childNode.ChildNodes;
					ArrayList temp_sequence = new ArrayList();
					foreach(XmlNode sequenceNode in SequenceNodes)//<Sequence>
					{
						XmlNodeList hopNodes = sequenceNode.ChildNodes; 

						foreach(XmlNode hopNode in hopNodes)//<Hop>
						{
							XmlNodeList pathNodes = hopNode.ChildNodes;
							Sequence s = new Sequence();
							foreach(XmlNode pathNode in pathNodes)
							{

								XmlElement cn = (XmlElement)pathNode;
								if(cn.Name.Equals("Destination"))
								{
									string temp = cn.FirstChild.Value;

									Point p = new Point();
									p.row = getRow(temp);
									p.col = getCol(temp);
									s.destination = p;
								}
								else if(cn.Name.Equals("Speed"))
								{
									s.speed = Convert.ToInt32(cn.FirstChild.Value);
								}
								else//Pause
								{
									s.pause = Convert.ToInt32(cn.FirstChild.Value);
								}
							}
							temp_sequence.Add(s);
						}
					}//outer foreach
					hashSequence.Add(source,temp_sequence);
				}									               
			}
			catch(Exception ex)
			{
				MessageBox.Show("Exception in storeSequenceInHashtable():"+ex.Message);
			}   
		}
 
		/// <summary>
		/// When string (row,col) is passed as a parameter, return row as an integer
		/// </summary>
		/// <param name="s">destination</param>
		/// <returns>row as an integer</returns>
		public int getRow(string s)
		{
			try
			{
				int leftP = s.IndexOf("(");		
				int dot = s.IndexOf(",");
				string strRow = s.Substring(leftP+1, dot-1);
				int row = Convert.ToInt32(strRow);
				return row;
			}
			catch(Exception e)
			{
				MessageBox.Show("getRow() Exception occurred :"+e.Message);                
				return -1;
			}
		}

		/// <summary>
		/// When string (row,col) is passed as a parameter, return col as an integer
		/// </summary>
		/// <param name="s">destination</param>
		/// <returns>col as an integer</returns>
		public int getCol(string s)
		{
			try
			{
				int rightP = s.IndexOf(")");
				int dot = s.IndexOf(",");
				int left = dot+1;
				int right =rightP-left;
				string strCol = s.Substring(dot+1, right);
				int col = Convert.ToInt32(strCol);
				return col;
			}
			catch(Exception e)
			{
				MessageBox.Show("getCol() Exception occurred :"+e.Message);                
				return -1;
			}
		}

		/// <summary>
		/// Return an integer for the randomly selected sequence number 
		/// between 0 and 100000
		/// </summary>
		/// <returns></returns>
		private int getSessionID()
		{
			Random random = new Random();
			int random_seuqnce_num = random.Next(0, 100000);
			return random_seuqnce_num;
		}

		/// <summary>
		/// Return a request ID as a format: "session id : initiator: target: sequenceNum" 
		/// </summary>
		/// <param name="p_target"></param>
		/// <returns></returns>
		private string getRequestID(string p_type)
		{
			string requestID = "";
			if(p_type.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
			{
				requestID = MySessionID+":"+"0";				
			}
			else if(p_type.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
			{
				requestID = MySessionID+":"+int.MaxValue.ToString();				
			}
			else if(p_type.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
			{
				discoverSeqNum--;  
				requestID = MySessionID+":"+discoverSeqNum;				
			}
			else if(p_type.Equals(Packet_CONSTANTS.DATA_PACKET))
			{
				sequenceNum++;
				requestID = MySessionID+":"+sequenceNum;  
			}
			else if(p_type.Equals(Packet_CONSTANTS.ROUTE_ERROR_PACKET))
			{
				REPSeqNum++;
				requestID = MySessionID+":"+REPSeqNum;
			}
			return requestID;		
		}

		//UdpServer related methods
		public void start_server()
		{	
			try
			{		
				//Starting the UDP Listening thread.
				udpListeningThread = new Thread(new ThreadStart(this.receivePacket));
				udpListeningThread.Start();
				udpServerIsAlive = true;				
			}
			catch (Exception e)
			{
				MessageBox.Show("start_server() exception occurred : "+e.Message);
				
			}
		}
		
		/// <summary>
		///	This function will stop  and close the UDP server.  And
		/// it also aborts the running threads  
		/// </summary>
		public void stop_server()
		{
			try
			{	
				if (udpSocket != null)
				{
					udpSocket.Close();
				}
				udpListeningThread=null;//Free Mem
			}
			catch (Exception e)
			{
				MessageBox.Show("stop_server() exception occurred :"+e.Message);
			}
		}

		/// <summary>
		/// Receive UDP Packets
		/// </summary>
		public void receivePacket()
		{			
			IPHostEntry localHostEntry;
		
			try
			{
				udpSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
				localHostEntry = Dns.GetHostByName(Dns.GetHostName());
				IPEndPoint localIpEndPoint = new IPEndPoint(localHostEntry.AddressList[0], UDPCONSTANTS.UDPPORT);
				EndPoint localEndEntry = (localIpEndPoint);
				udpSocket.Bind(localEndEntry);
					
				while (udpServerIsAlive)
				{
					Monitor.Enter(this);
					Byte[] received = new Byte[UDPCONSTANTS.MESSAGESIZE];

					//This is just a dummy endpoint that will be replaced after the call to ReceiveFrom.
					//In the ReceiveFrom call later on, dummyIpEndPoint will be passed by ref.
					IPEndPoint dummyIpEndPoint = new IPEndPoint(localHostEntry.AddressList[0], UDPCONSTANTS.UDPPORT);

					EndPoint remoteNotifyEP = (dummyIpEndPoint);

					int udpBytesReceived = udpSocket.ReceiveFrom(received, ref remoteNotifyEP);

					ReceivedDataLength=udpBytesReceived;								
					String dataReceivedUdp = System.Text.Encoding.ASCII.GetString(received,0,received.Length );
					ReceivedData=dataReceivedUdp.ToString();
					//HandleReceivedPacket();
					controlInvoker.Invoke(new MethodCallInvoker(HandleReceivedPacket));					
					Monitor.Exit(this);					
				}		
					
			}	
			catch (Exception e)
			{
				//MessageBox.Show("Exception occurred in receivedPacket() :"+e.Message);
			}
				
		}

		//UdpClient relate methods
		public bool sendUDP(string serverIPaddr, string message)
		{
			bool sentSuccess = false;
			try
			{
				IPAddress serverIP = IPAddress.Parse(serverIPaddr);
				udpClient = new UdpClient();
				udpClient.Connect(serverIP,UDPCONSTANTS.UDPPORT);
		
				Byte[] inputToBeSent = new Byte[UDPCONSTANTS.MESSAGESIZE];
				//Convert the "message" string to a byte array, and then send it to
				//the remote UDP server using udpClient.
				inputToBeSent = System.Text.Encoding.ASCII.GetBytes(message.ToCharArray());
				int nBytesSent = udpClient.Send(inputToBeSent, inputToBeSent.Length);
				sentSuccess = true;
				
			}
			catch(SocketException se)
			{
				//MessageBox.Show("Exception in sendUDP(): SocketException: " +se.Message);
			}
			catch(Exception e)
			{
				/*MessageBox.Show("Exception is occurred in sendUDP() : "+e.Message+"\r\n serverIP: "+
					serverIPaddr+" Message: "+message);
				MessageBox.Show("Inner Exception in sendUDP(): "+e.InnerException);*/
			}
			udpClient.Close();
			return sentSuccess;
		}

		private Packet getPacketFromReceiptBuffer(string p_receiptID)
		{
			Packet packet = new Packet();
			IDictionaryEnumerator ide = ReceiptBuffer.GetEnumerator();
			while(ide.MoveNext())
			{
				if(ide.Key.ToString().Equals(p_receiptID))
				{
					packet = (Packet)ide.Value;
					break;
				}
			}
			return packet; 
		}
		/// <summary>
		/// Send a route error packet to the initiator.
		/// It does not have route reply packet back
		/// If source does not have any packet such as route reply packet 
		/// or route error packet, it will time out.
		/// </summary>
		public void InformRouteErrorToSource(Packet p_packet)
		{
			try
			{			
				if(p_packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
					p_packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET) ||
					p_packet.strType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET) ||
					p_packet.strType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET))
				{
					string p_receiptID = p_packet.strReceiptID;
					if(ReceiptBuffer.Contains(p_receiptID))
					{
						Packet packet = getPacketFromReceiptBuffer(p_receiptID);
						ReceiptBuffer.Remove(p_receiptID);
						SendRouteErrorPacket(packet);
					}
				}
			}//try
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in InformRouteErrorToSource() : "+e.Message);
			}

		}


		/// <summary>
		/// Send route error packet to the source.
		/// </summary>
		/// <param name="p_packet">packet which is not transmitted: got it from receipt buffer</param>
		private void SendRouteErrorPacket(Packet p_packet)
		{
			try
			{
				string originalType = p_packet.strType;
				string temp_RouteRecord = p_packet.strRouteRecord;
				p_packet.strType = Packet_CONSTANTS.ROUTE_ERROR_PACKET;                
				//string errorPacketID = p_packet.strRequestID;
				string source = "";

				if(originalType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
					originalType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
				{
					source = p_packet.strInitiator;
				}
				else if(originalType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET) ||
					originalType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET))//reply packets
				{
					source = p_packet.strTarget;
				}
				p_packet.strInitiator = strMyID;
				p_packet.strTarget = source;
				p_packet.strRouteRecord = strMyID;
				p_packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				p_packet.strCurrentID = strMyID;
				p_packet.strRouteCache = getReverseRoute(temp_RouteRecord);
				if(source.Equals(strMyID))//when route error ocurrs in source itself
				{
					string serverIP = GetIPAddressByIDInConfFile(source);					
					SaveInReceiptBuffer(p_packet);//for a receipt packet
					string messageToSend = CreateSendingMessage(p_packet);
					ResultPacket rp = new ResultPacket();
					rp.strComType = Packet_CONSTANTS.SENT;
					storeResultPackets(rp, p_packet);
					sendUDP(serverIP, messageToSend);
				}
				else
				{
					SendPacket(ref p_packet);				
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendRouteErrorPacket(): "+e.Message);
			}		
		}

		private string CreateSendingMessage(Packet p_packet)
		{
			string messageToSend = "NA";
			try
			{
				XmlDocument doc = new XmlDocument();
				doc.LoadXml("<Packet></Packet>");

				XmlNode root = doc.DocumentElement;

				//1. Type
				XmlElement elem = doc.CreateElement(Packet_CONSTANTS.TYPE);
				elem.InnerText = p_packet.strType;
				root.AppendChild(elem);


				//2.Initiator
				elem = doc.CreateElement(Packet_CONSTANTS.INITIATOR);
				elem.InnerText = p_packet.strInitiator;
				root.AppendChild(elem);


				//3.Current ID
				elem = doc.CreateElement(Packet_CONSTANTS.CURRENTID);
				elem.InnerText = p_packet.strCurrentID;
				root.AppendChild(elem);
				
				//4.Payload
				elem = doc.CreateElement(Packet_CONSTANTS.PAYLOAD);
				elem.InnerText = p_packet.strPayload;
				root.AppendChild(elem);

				//5. RequestID
				elem = doc.CreateElement(Packet_CONSTANTS.REQUESTID);
				elem.InnerText = p_packet.strRequestID;
				root.AppendChild(elem);

				//6.RouteCache
				elem = doc.CreateElement(Packet_CONSTANTS.ROUTECACHE);
				elem.InnerText = p_packet.strRouteCache;
				root.AppendChild(elem);

				//7.RouteRecord
				elem = doc.CreateElement(Packet_CONSTANTS.ROUTERECORD);
				elem.InnerText = p_packet.strRouteRecord;
				root.AppendChild(elem);

				//8.Target
				elem = doc.CreateElement(Packet_CONSTANTS.TARGET);
				elem.InnerText = p_packet.strTarget;
				root.AppendChild(elem);
				
				//9.ReceiptID
				elem = doc.CreateElement(Packet_CONSTANTS.RECEIPTID);
				elem.InnerText = p_packet.strReceiptID;
				root.AppendChild(elem);

				//10. Put TimeStamp into packet
				elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
				p_packet.strTimeStamp = getNTTime().ToString();
				elem.InnerText = p_packet.strTimeStamp;
				root.AppendChild(elem);
				messageToSend = doc.InnerXml.ToString();
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in CreateSendingMessage(): "+e.Message);
			}
			return messageToSend;
		}

		public Packet getLostPacketFromSendBuffer(string p_requestID)
		{
			Packet packet = new Packet();
			try
			{
				IDictionaryEnumerator ide = send_buffer.GetEnumerator();
				while(ide.MoveNext())
				{
					if(p_requestID.Equals(ide.Key.ToString()))
					{
						packet = (Packet)ide.Value;
						break;
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getLostPacketFromSendBuffer():"+e.Message);
			}
			return packet;
		}

		/// <summary>
		/// this method is performed in the source node, which has received route error packet.
		/// </summary>
		/// <param name="p_packet"></param>
		public void PerformRouteDiscovery(string p_REPRID, Packet p_packet)
		{
			try
			{
				if(send_buffer.Contains(p_packet.strRequestID) && 
					!processingRDBuffer.Contains(p_packet.strRequestID))
				{
					Packet packet = new Packet();
					processingRDBuffer.Add(p_packet.strRequestID);
					packet.strType = Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET;
					packet.strInitiator = strMyID; //which is source itself
					packet.strTarget = p_packet.strTarget;
					if(p_REPRID.Equals(Packet_CONSTANTS.NOTAVAILABLE))
					{
						packet.strRequestID = getRequestID(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET);
					}
					else
					{
                        packet.strRequestID = p_REPRID;
					}
					packet.strCurrentID = strMyID;
					//by using payload slot, put the lost packet's request id.
					packet.strPayload = p_packet.strRequestID;
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
					packet.strRouteRecord = strMyID;				
					SaveInSendBuffer(packet);
					string timerID = packet.strRequestID;
					if(!RRTimer2Buffer.Contains(timerID))
					{
						SetRouteReplyTimer(timerID, packet);
						SendPacket(ref packet);		
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in PerformRouteDiscovery():"+e.Message);
			}
		}

		/// <summary>
		/// This repairing process should be done in source node.
		/// If route are available in source's cache, source uses it. 
		/// Otherwise, perform route discovery.
		/// </summary>
		/// <param name="packet"></param>
		public void RepairRouteError(string p_RID, ref Packet packet)
		{
			try
			{
				if(send_buffer.Contains(packet.strRequestID))
				{
					string requestID = packet.strRequestID;
					Packet sb_packet = getLostPacketFromSendBuffer(requestID);
					if(sb_packet.strInitiator.Equals(strMyID))
					{
						if(!routeCache.isEmpty())
						{
							processingRDBuffer.Remove(sb_packet.strRequestID);
							sb_packet.strRouteCache = routeCache.getPath();
							string requestIDSB = sb_packet.strRequestID;
							send_buffer.Remove(requestIDSB);
							send_buffer.Add(requestIDSB, sb_packet);
							if(!RRTimer2Buffer.Contains(sb_packet.strRequestID))
							{
								SetRouteReplyTimer(sb_packet.strRequestID, sb_packet);
								SendPacket(ref sb_packet);		
							}
						}
						else
						{
							PerformRouteDiscovery(p_RID, sb_packet);
						}                 
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception ocurred in RepairRouteError():"+e.Message);
			}
		}

		private string CreateBroadcastPacket(BroadcastPacket p_packet)
		{
			string messageToSend ="";
			try
			{
				XmlDocument doc = new XmlDocument();
				doc.LoadXml("<Packet></Packet>");

				XmlNode root = doc.DocumentElement;

				//1. Type
				XmlElement elem = doc.CreateElement(Packet_CONSTANTS.TYPE);
				elem.InnerText = p_packet.strType;
				root.AppendChild(elem);
				//2. Cache on or off
				elem = doc.CreateElement(Packet_CONSTANTS.CACHE);
				elem.InnerText = p_packet.strOnCache;
				root.AppendChild(elem);				
				//3. ACM on or off
				elem = doc.CreateElement(Packet_CONSTANTS.ACM);
				elem.InnerText = p_packet.strOnACM;
				root.AppendChild(elem);				
				//4. SWS
				elem = doc.CreateElement(Packet_CONSTANTS.SWS);
				elem.InnerText = p_packet.strSWS;
				root.AppendChild(elem);								
				//5. Number of Data Packet
				elem = doc.CreateElement(Packet_CONSTANTS.NUMOFDATAPACKET);
				elem.InnerText = p_packet.strNumOfDP;
				root.AppendChild(elem);								
				//6. Use of Receipt Packet
				elem = doc.CreateElement(Packet_CONSTANTS.RP);
				elem.InnerText = p_packet.strOnRP;
				root.AppendChild(elem);				
				messageToSend = doc.InnerXml.ToString();				
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in CreateBroadcastPacket(): "+e.Message);
			}
			return messageToSend;

		}
		private void SendBroadcastNTPacket()
		{
			try
			{

				BroadcastPacket broadPacket = new BroadcastPacket();
				broadPacket.strType = Packet_CONSTANTS.BROADCAST_NT_PACKET;				
				broadPacket.strOnACM = ACM_CB.Checked.ToString();
				broadPacket.strOnCache = Cache_CB.Checked.ToString();
				broadPacket.strSWS = SWS_TB.Text;
				broadPacket.strNumOfDP = NumOfData_TB.Text;
				broadPacket.strOnRP = RP_CB.Checked.ToString();
				
				string broadcastNT = CreateBroadcastPacket(broadPacket);
				IDictionaryEnumerator ide = hashBuddyList.GetEnumerator();
				while(ide.MoveNext())
				{
					string serverID = ide.Key.ToString();					
					string serverIP = GetIPAddressByIDInConfFile(serverID);
					sendUDP(serverIP, broadcastNT);
				}	
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in BroadcastNTPacket() : "+e.Message);
			}
		}

		/// <summary>
		/// Set ACM on or off
		/// </summary>
		private void SetACM()
		{
			try
			{
				if(ACM_CB.Checked)
				{
					OnACM = true;
				}
				else
				{
					OnACM = false;
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SetACM(): "+e.Message);
			}
		}

		private void SetCache()
		{
			try
			{
				if(Cache_CB.Checked)
				{
					OnCache = true;
				}
				else
				{
					OnCache = false;
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SetCache(): "+e.Message);
			}
		}

		private void buddy_Btn_Click(object sender, System.EventArgs e)
		{	
			try
			{					
				//MessageBox.Show(getNTTime().ToString());
				Random r = new Random();
				int RandomNum = r.Next(1,hashBuddyList.Count+1);
				int i = 0;
				IDictionaryEnumerator ide = hashBuddyList.GetEnumerator();
				while(ide.MoveNext())
				{
					i++;
					if(ide.Key.ToString().Equals(strMyID) && 
						(i.ToString().Equals(RandomNum.ToString())))
					{
						ide.Reset();
						i =0;
						RandomNum = r.Next(1, hashBuddyList.Count+1);
					}
					else if(i.ToString().Equals(RandomNum.ToString()))
					{
                        strBuddyID = ide.Key.ToString();
						strBuddyIP = ide.Value.ToString();
						break;
					}
				}
				BuddyID_IP.Text = strBuddyID+" : "+strBuddyIP;
			}
			catch(Exception ex)
			{
                MessageBox.Show("Exception in buddy_Btn_Click(): "+ex.Message);
			}
		}

		private void writeBasicInfoXML()
		{
			try
			{
				XmlTextWriter basicInfoWriter = 
					new XmlTextWriter(NT_CONSTANTS.BASIC_INFO_FILENAME, Encoding.ASCII);
				basicInfoWriter.Formatting = Formatting.Indented;
				basicInfoWriter.WriteStartDocument();
				basicInfoWriter.WriteStartElement("ADC_BasicInfo");//root <ADC_BasicInfo>
				string sourceID_IP = strMyID+":"+strMyIP;
				string targetID_IP = strBuddyID+":"+strBuddyIP;
				basicInfoWriter.WriteElementString("Source",sourceID_IP);				
				basicInfoWriter.WriteElementString("Target",targetID_IP);
				basicInfoWriter.WriteElementString("ACM",OnACM.ToString());
				basicInfoWriter.WriteElementString("Cache",OnCache.ToString());
				basicInfoWriter.WriteElementString("SWS",SWS.ToString());
				basicInfoWriter.WriteElementString("NumOfDataPacket",NumOfDataPacket.ToString());
				basicInfoWriter.WriteElementString("ReceiptPacket",OnReceiptPacket.ToString());
				basicInfoWriter.WriteEndElement();//</ADC_BasicInfo>
                basicInfoWriter.WriteEndDocument();
				basicInfoWriter.Close();
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in writeBasicInfoXML(): "+e.Message);
			}
		}

		private void Start_Btn_Click(object sender, System.EventArgs e)
		{
			if(!BuddyID_IP.Text.Equals(""))
			{	
				SendBroadcastNTPacket();
			}
			else
			{
                MessageBox.Show("Buddy is not selected!");
			}
		}


		private void Reset_Btn_Click(object sender, System.EventArgs e)
		{
			try
			{
				NumOfData_TB.Text ="";
				SWS_TB.Text ="";
				Cache_CB.Checked = false;
				RP_CB.Checked = false;
				ACM_CB.Checked = false;
			}
			catch(Exception ex)
			{
                MessageBox.Show("Exception in Reset_Btn_Click(): "+ex.Message);
			}			
		}
      
		private ResultPacket getLastRRP()
		{   
			ResultPacket rp = new ResultPacket();
			try
			{
				for(int i = resultList.Count-1; i>0 ; i--)
				{
					ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("RRP"))
					{
						rp = tempRP;
						break;
					}
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getLastRRP(): "+e.Message);
			}
			return rp;
		}
		private string getThroughput()
		{
			string throughput = "N/A";
			try
			{
				ResultPacket tempRPForBT = (ResultPacket)resultList[0];
				double BT = double.Parse(tempRPForBT.strTimeStampInSender);
				ResultPacket tempRPForET = getLastRRP();
				double ET = double.Parse(tempRPForET.strTimeStampInReceiver);
				double TDPD = double.Parse(NumOfDataPacket.ToString());//total data packet delivered
				double doubleTP = TDPD / (ET-BT);
                double decimal2TP =  Math.Round(doubleTP, 2);
				throughput = decimal2TP.ToString();
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getThroughput(): "+e.Message);
			}
			return throughput;
		}

		private bool isSourceNode()
		{
			try
			{
				if(File.Exists(NT_CONSTANTS.BASIC_INFO_FILENAME))
				{
                    return true;
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in isSourceNode(): "+e.Message);
			}
			return false;
		}

		private bool isTargetNode()
		{
			try
			{
				string tempID_IP = strBuddyID+" : "+strBuddyIP;			
				StreamReader myReader = new StreamReader(NT_CONSTANTS.BASIC_INFO_FILENAME);
				string fileContent = myReader.ReadToEnd();
				XmlDocument xmlDoc = new XmlDocument();
				xmlDoc.LoadXml(fileContent);
				XmlNode root = xmlDoc.DocumentElement;
				XmlNodeList childNodes = root.ChildNodes;
				foreach(XmlNode childNode in childNodes)
				{
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals("Target"))
					{
						if(elem.Value.Equals(tempID_IP))
						{
							return true;
						}
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in isTargetNode(): "+e.Message);
			}
			return false;
		}

		private string getTotalNumOfCPSent()
		{
			int intNCP = 0;
			try
			{
				for(int i =0; i< resultList.Count; i++)
				{ 
					ResultPacket rp = (ResultPacket)resultList[i];
					if(rp.strComType.Equals(Packet_CONSTANTS.SENT) &&
						!rp.strType.Equals("DP"))
					{
						intNCP++;
					}
				}				
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getNormalizedControlOverhead(): "+e.Message);
			}
			return intNCP.ToString() ;
		}

		private Hashtable getDPs()
		{
            Hashtable DPHash = new Hashtable();
	        try
			{
				for(int i = 0; i<resultList.Count; i++)
				{
                    ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("DP"))
					{
                        string tempRID = tempRP.strRequestID;
						if(!DPHash.Contains(tempRID))
						{
                            string TSInSender = tempRP.strTimeStampInSender;
							DPHash.Add(tempRID, TSInSender);
						}
  					}
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getDPs(): "+e.Message);
			}
			return DPHash;
		}

		private Hashtable getRRPs()
		{
			Hashtable RRPHash = new Hashtable();
			try
			{
				for(int i = 0; i<resultList.Count; i++)
				{
					ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("RRP"))
					{
						string tempRID = tempRP.strRequestID;
						string TSInSender = tempRP.strTimeStampInSender;
						RRPHash.Add(tempRID, TSInSender);
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getDPs(): "+e.Message);
			}
			return RRPHash;
		}

		private string getTSInRRPs(Hashtable p_hash, string p_id)
		{
            string TS = "N/A";
			try
			{
				IDictionaryEnumerator ide = p_hash.GetEnumerator();
				while(ide.MoveNext())
				{
                    string tempRID = ide.Key.ToString();
					if(tempRID.Equals(p_id))
					{
                        TS = ide.Value.ToString();
						break;
					}
  				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getTSInRRPs(): "+e.Message);
			}
			return TS;
		}
		private string getAvgLatencyForDeliveryingDP()
		{
			string strAvgL = "N/A";
			try
			{
				ArrayList ALDDPList = new ArrayList();
				Hashtable DPHash = getDPs();
				Hashtable RRPHash = getRRPs();
				IDictionaryEnumerator ide = DPHash.GetEnumerator();
				while(ide.MoveNext())
				{
                    string tempRIDInDPs = ide.Key.ToString();
					if(RRPHash.Contains(tempRIDInDPs))
					{
                        int DP_TSInSender = int.Parse(ide.Value.ToString());
						int RRP_TSInSender = int.Parse(getTSInRRPs(RRPHash, tempRIDInDPs));
						int D = RRP_TSInSender - DP_TSInSender;
						ALDDPList.Add(D.ToString());
					}
				}
				double AvgL = getAvg(ALDDPList);
				strAvgL = AvgL.ToString();
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getAvgLatencyForDeliveryingDP(): "+e.Message);
			}
			return strAvgL;
		}

		private double getAvg(ArrayList p_avgTime)
		{
			double avg = 0;
			double sum = 0;
			try
			{
				IEnumerator ie = p_avgTime.GetEnumerator();
				while(ie.MoveNext())
				{
					sum += double.Parse(ie.Current.ToString());
				}
				avg = (sum / double.Parse(p_avgTime.Count.ToString()));
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getAvg(): "+e.Message);
			}
			double decimal2Avg = Math.Round(avg, 2);
			return decimal2Avg;
		}

		private int getNumOfDPSent()
		{
			int numOfDPSent = 0;
			try
			{
				for(int i =0; i<resultList.Count; i++)
				{
                    ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("DP"))
					{
                       numOfDPSent++;
					}
  				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getNumOfDPSent() : "+e.Message);
			}
			return numOfDPSent;
		}
		
		private string getDeliveryRatioOfDP()
		{
			double DRDP = 0;
			try
			{
				double numOfDPSent = double.Parse(getNumOfDPSent().ToString());
				double totalNumOfDPDelivered = double.Parse(NumOfDataPacket.ToString());
				DRDP = (totalNumOfDPDelivered / numOfDPSent);
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getDeliveryRatioOfDP() : "+e.Message);
			}
			double decimal2DRDP = Math.Round(DRDP, 2);
			return decimal2DRDP.ToString();
		}

		private void writeToFileHash(Hashtable p_hash, string p_file_name)
		{
			try
			{
				if(isSourceNode())
				{
					string path = "\\My Documents\\Personal\\"+p_file_name;
					StreamWriter myWriter = new StreamWriter(path);
					IDictionaryEnumerator ide = p_hash.GetEnumerator();
					while(ide.MoveNext())
					{
						myWriter.WriteLine(ide.Key.ToString()+" "+ide.Value.ToString());
					}
					myWriter.Close();
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in writeToFileHash(): "+e.Message);
			}
		}
		private Hashtable getREPHash()
		{
			Hashtable REPHash = new Hashtable();
			try
			{
				for(int i= 0; i<resultList.Count; i++)
				{
					ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("REP") && 
						tempRP.strComType.Equals(Packet_CONSTANTS.RECEIVED))
					{
						string tempRID = tempRP.strRequestID;
						if(!REPHash.Contains(tempRID))
						{
							REPHash.Add(tempRID, tempRP.strTimeStampInReceiver);
						}
					}
				}
				writeToFileHash(REPHash, "REPHash.txt");
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getREPHash(): "+e.Message);
			}
			return REPHash;
		}

		private Hashtable getRDPHash()
		{
			Hashtable RDPHash = new Hashtable();
			try
			{
				for(int i= 0; i<resultList.Count; i++)
				{
					ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("RDP"))
					{
						string tempRID = tempRP.strRequestID;
						if(!RDPHash.Contains(tempRID))
						{
							RDPHash.Add(tempRID, tempRP.strTimeStampInSender);
						}
					}
				}
                writeToFileHash(RDPHash, "RDPHash.txt"); 
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getRDPHash(): "+e.Message);
			}
			return RDPHash;

		}
		private Hashtable getRRDPHash()
		{
			Hashtable RRDPHash = new Hashtable();
			try
			{
				for(int i= 0; i<resultList.Count; i++)
				{
					ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("RRDP") &&
						!RRDPHash.Contains(tempRP))
					{
						string tempRID = tempRP.strRequestID;
						RRDPHash.Add(tempRID, tempRP.strTimeStampInReceiver);
					}
				}
				writeToFileHash(RRDPHash, "RRDPHash.txt");
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getRRDPHash(): "+e.Message);
			}
			return RRDPHash;
		}

		private Hashtable getRRPHash()
		{
            Hashtable RRPHash = new Hashtable();
			try
			{
				for(int i= 0; i<resultList.Count; i++)
				{
					ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("RRP"))
					{
						string tempRID = tempRP.strRequestID;
						RRPHash.Add(tempRID, tempRP.strTimeStampInReceiver);
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getRRPHash(): "+e.Message);
			}
			return RRPHash;
		}

		private int getConnectionEstablishmentTime()
		{
            int CE = 0;
			try
			{
                ResultPacket tempRP_ST = (ResultPacket)resultList[0];
				int ST = int.Parse(tempRP_ST.strTimeStampInSender);
				for(int i = 0 ; i<resultList.Count; i++)
				{
                    ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("ASCP"))
					{
                        int ET = int.Parse(tempRP.strTimeStampInReceiver);
						CE = ET - ST;
						break;
 					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getConnectionEstablishmentTime(): "+e.Message);
			}
			return CE;
		}

		private int getTime(Hashtable p_hash, string p_id)
		{
            int T = 0;
			try
			{
                IDictionaryEnumerator ide = p_hash.GetEnumerator();
				while(ide.MoveNext())
				{
					if(ide.Key.ToString().Equals(p_id))
					{
                        T = int.Parse(ide.Value.ToString());
						break;
					}
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getTime(): "+e.Message);
			}
			return T;
		}

		private string getNotTransmittedDPRIDFromREPs(string p_id)
		{
			string RID = "N/A";
			try
			{
				for(int i = 0; i< resultList.Count; i++)
				{
                    ResultPacket tempRP = (ResultPacket)resultList[i];
					if(tempRP.strType.Equals("REP") &&
						tempRP.strComType.Equals(Packet_CONSTANTS.RECEIVED))
					{
						if(tempRP.strRequestID.Equals(p_id))
						{
                            RID = tempRP.strNotTransmittedDP;
							break;
						}
					}
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getNotTransmittedDPRIDFromREPs(): "+e.Message);
			}
			return RID;
		}
		private string getAvgLatencyForFindingNewRoute()
		{
			string strAvg = "N/A";
			try
			{
				Hashtable REPHash = getREPHash();
				Hashtable tempREPHash = getREPHash();
				Hashtable RDPHash = getRDPHash();
				Hashtable RRDPHash = getRRDPHash();
				Hashtable RRPHash = getRRPHash();
				ArrayList Latencies = new ArrayList();

				IDictionaryEnumerator ide = REPHash.GetEnumerator();
				while(ide.MoveNext())
				{
                    string tempRID = ide.Key.ToString();
					int ST = int.Parse(ide.Value.ToString());
					if(RRDPHash.ContainsKey(tempRID))
					{
                        int ET = getTime(RRDPHash, tempRID);
						int L = ET - ST;
						if(L>0)
						{
							Latencies.Add(L.ToString());
						}
						if(RDPHash.Contains(tempRID))
						{
							RDPHash.Remove(tempRID);
							tempREPHash.Remove(tempRID);
						}
					}
				}

				IDictionaryEnumerator ide_RDP = RDPHash.GetEnumerator();
				while(ide_RDP.MoveNext())
				{
                    string tempRDPRID = ide_RDP.Key.ToString();
					int ST = int.Parse(ide_RDP.Value.ToString());
					if(RRDPHash.Contains(tempRDPRID))
					{
						int ET = getTime(RRDPHash, tempRDPRID);
						int L = ET - ST;
						if(L>0)
						{
							Latencies.Add(L.ToString());
						}
					}
				}
				if(tempREPHash.Count != 0)
				{
					IDictionaryEnumerator ide_REP = tempREPHash.GetEnumerator();
					while(ide_REP.MoveNext())
					{
					    string ID = ide_REP.Key.ToString();
						string NotTramsmittedDPRID = getNotTransmittedDPRIDFromREPs(ID);
						int ST = int.Parse(ide_REP.Value.ToString());
						if(RRPHash.Contains(NotTramsmittedDPRID))
						{
							int ET = getTime(RRPHash, NotTramsmittedDPRID);
							int L = ET - ST;
							if(L>0)
							{
								Latencies.Add(L.ToString());
							}                           
						}
					}
				}
				int connectionTime = getConnectionEstablishmentTime();
				Latencies.Add(connectionTime.ToString());
				double avg = getAvg(Latencies);
				strAvg = avg.ToString();
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getAvgLatencyForFindingNewRoute(): "+e.Message);
			}
			return strAvg;
		}

		private string getAvgNumOfCacheEntries()
		{
			string strAvgCacheEntries = "N/A";
			try
			{
                ArrayList CacheList = new ArrayList();
				for(int i =0; i<resultList.Count; i++)
				{
                    ResultPacket tempRP = (ResultPacket)resultList[i];
					string numOfRoutesInCache = tempRP.strNumOfRoutesInCache;
					CacheList.Add(numOfRoutesInCache);
				}
				double avg = getAvg(CacheList);
                strAvgCacheEntries = avg.ToString();
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getAvgNumOfCacheEntries(): "+e.Message);
			}
			return strAvgCacheEntries;
		}

		private string getSpeed()
		{
			string speed = "N/A";
			try
			{
				string path = "\\My Documents\\Personal\\GEN_basicInfo.xml";
				StreamReader myReader = new StreamReader(path, Encoding.ASCII);
				string fileContent = myReader.ReadToEnd();
				XmlDocument xmlDoc = new XmlDocument();
				xmlDoc.LoadXml(fileContent);
				XmlNode root = xmlDoc.DocumentElement;
				XmlNodeList childNodes = root.ChildNodes;
				foreach(XmlNode childNode in childNodes)
				{
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals("Speed"))
					{
						speed = elem.FirstChild.Value;
						//MessageBox.Show("Speed: "+speed);
						break;
					}
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in getSpeed(): "+e.Message);
			}
			return speed;
		}
		private string getPauseTime()
		{
			string PT = "N/A";
			try
			{
				string path = "\\My Documents\\Personal\\GEN_basicInfo.xml";
				StreamReader myReader = new StreamReader(path, Encoding.ASCII);
				string fileContent = myReader.ReadToEnd();
				XmlDocument xmlDoc = new XmlDocument();
				xmlDoc.LoadXml(fileContent);
				XmlNode root = xmlDoc.DocumentElement;
				XmlNodeList childNodes = root.ChildNodes;
				foreach(XmlNode childNode in childNodes)
				{
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals("PauseTime"))
					{
						PT = elem.FirstChild.Value;
						//MessageBox.Show("PT: "+PT);
						break;
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getPauseTime(): "+e.Message);
			}
			return PT;
		}
		private void writeDataAnalysisXML()
		{
			try
			{
				string file_path = "\\My Documents\\Personal\\DataAnalysis_"+strMyID+".xml";
			    XmlTextWriter myWriter = new XmlTextWriter(file_path,Encoding.ASCII);
				myWriter.Formatting = Formatting.Indented;
				myWriter.WriteStartDocument();
				myWriter.WriteStartElement("DataAnalysis");
				myWriter.WriteAttributeString("ID", strMyID);
				if(isSourceNode())
				{
					myWriter.WriteAttributeString("SourceNode", "Yes");
				}
				else
				{
					myWriter.WriteAttributeString("SourceNode", "No");
				}
				myWriter.WriteAttributeString("TimeNow", getNTTime().ToString());
				if(isSourceNode())
				{
                    myWriter.WriteStartElement("XParam");
					string tempSpeed = getSpeed();
					string tempPT = getPauseTime();
					myWriter.WriteElementString("SWS", SWS.ToString());
					myWriter.WriteElementString("Speed", tempSpeed);
					myWriter.WriteElementString("PauseTime", tempPT);
					myWriter.WriteEndElement();				
					string throughput = getThroughput();				
					string TotalDataPacketDelivered = NumOfDataPacket.ToString();
					string AvgLatencyDeliveryingDP = getAvgLatencyForDeliveryingDP();
					string AvgLatencyFindingNewRoute = getAvgLatencyForFindingNewRoute();
					string DeliveryRatioOfDP = getDeliveryRatioOfDP();
					string AvgNumOfCacheEntries = getAvgNumOfCacheEntries();
					

					myWriter.WriteElementString("Throughput", throughput);
					myWriter.WriteElementString("TotalDataPacketDelivered", TotalDataPacketDelivered);
					myWriter.WriteElementString("AvgLatencyDeliveryingDP", AvgLatencyDeliveryingDP);
					myWriter.WriteElementString("AvgLatencyFindingNewRoute", AvgLatencyFindingNewRoute);
					myWriter.WriteElementString("DeliveryRatioOfDP", DeliveryRatioOfDP);
					myWriter.WriteElementString("AvgNumOfCaCheEntries", AvgNumOfCacheEntries);
				}
				string TotalNumberOfCPSent = getTotalNumOfCPSent();
				myWriter.WriteElementString("TotalNumberOfCPSent", TotalNumberOfCPSent);
				myWriter.WriteEndElement();
				myWriter.WriteEndDocument();
				myWriter.Close();
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in writeDataAnalysisXML(): "+e.Message);
			}
		}

		/*****************************************************************************/
		
		private void AddToReceiverWindow(Packet p_packet)
		{
			try
			{
				if(ReceiverWindow.Count<SWS)
				{
					int SN = getSequenceNumFromRequestIDReceived(p_packet.strRequestID);
					ElementsForReceiverWindow ERW = new ElementsForReceiverWindow();
					ERW.SN = SN.ToString();
					ERW.Payload = p_packet.strPayload;
					InsertInProperPositionInRW(ERW);
				}
					
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in AddToReceiverWindow(): "+e.Message);
			}
		}

		private void InsertInProperPositionInRW(ElementsForReceiverWindow p_ERW)
		{
			try
			{
				int receivedSN = int.Parse(p_ERW.SN);
				bool isInserted = false;
				for(int i =0; i<ReceiverWindow.Count; i++)
				{
					ElementsForReceiverWindow ERW = (ElementsForReceiverWindow)ReceiverWindow[i];
					int tempSN = int.Parse(ERW.SN);
					if(receivedSN<tempSN)
					{					
						ReceiverWindow.Insert(i,p_ERW);
						isInserted = true;
						break;
					}
				}
				if(!isInserted)
				{
					ReceiverWindow.Add(p_ERW);
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in InsertInProperPositionInRW(): "+e.Message);
			}
		}
		private void UpdateSenderWindow(string p_RID)
		{
			try
			{
				if(SenderWindow.Contains(p_RID))
				{
					for(int i =0; i<SenderWindow.Count; i++)
					{
						if(SenderWindow[i].ToString().Equals(p_RID))
						{
							if(i==0)
							{
								SenderWindow.RemoveAt(0);
								while((SenderWindow.Count>0) &&
									SenderWindow[0].ToString().Equals("R"))
								{
                                    SenderWindow.RemoveAt(0);
								}
								break;
							}
							else 
							{
								SenderWindow.Insert(i,"R");
								SenderWindow.RemoveAt(i+1);
								break;
							}
						}						
					}
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in UpdateSenderWindow(): "+e.Message);
			}
		}

		private bool isInReceiverWindow(string p_SN)
		{
			bool isHere = false;
			try
			{
				IEnumerator ie = ReceiverWindow.GetEnumerator();
				while(ie.MoveNext())
				{
					ElementsForReceiverWindow ERW = (ElementsForReceiverWindow)ie.Current;
					if(ERW.SN.Equals(p_SN))
					{
                        isHere = true;
						break;
					}
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in isInReceiverWindow(): "+e.Message);
			}
			return isHere;
		}

		private void RP_CB_CheckStateChanged(object sender, System.EventArgs e)
		{
			OnReceiptPacket = RP_CB.Checked;
		}

		public void ContinueTryingAfterThreeRDPs(Packet p_packet)
		{
			try
			{
				if(send_buffer.Contains(p_packet.strRequestID) &&
					send_buffer.Contains(p_packet.strPayload))
				{
					if(routeCache.isEmpty())
					{
						SetRouteReplyTimer(p_packet.strRequestID, p_packet);
						SendPacket(ref p_packet);
					}
					else
					{
						send_buffer.Remove(p_packet.strRequestID);
						processingRDBuffer.Remove(p_packet.strPayload);
						if(send_buffer.Contains(p_packet.strPayload))
						{
							Packet lostPacket = getLostPacketFromSendBuffer(p_packet.strPayload);
							lostPacket.strRouteCache = routeCache.getPath();
							UpdateSendBuffer(lostPacket.strRequestID);
							send_buffer.Add(lostPacket.strRequestID, lostPacket);
							if(!RRTimer2Buffer.Contains(lostPacket.strRequestID))
							{
								SetRouteReplyTimer(lostPacket.strRequestID, lostPacket);
								SendPacket(ref lostPacket);
							}
						}
					}
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in ContinueTryingAfterThreeRDPs(): "+e.Message);
			}
		}

		private void SWS_TB_TextChanged(object sender, System.EventArgs e)
		{
		
		}
	}//main class
	
	public struct ElementsForReceiverWindow
	{
        public string SN;
		public string Payload;
	}
	
	/// <summary>
	/// This is a struct for Packet
	/// </summary>
	public struct Packet
	{
		public string strType;
		public string strInitiator;
		public string strTarget;
		public string strRequestID;
		public string strRouteRecord;
		public string strCurrentID;
		public string strPayload;
		public string strRouteCache;
		public string strReceiptID;
		public string strTimeStamp;

		public bool isDataPacket()
		{
			if(strType.Equals(Packet_CONSTANTS.DATA_PACKET))
			{				
				return true;
			}
			else
			{
				return false;
			}
		}
	}

	public struct ReceiptPacket
	{
		public string strType;
		public string strReceiptID;
		public string strTimeStamp;
	}

	public struct BroadcastPacket
	{
        public string strType;		
		public string strOnCache;
		public string strOnACM;
		public string strSWS;
		public string strNumOfDP;
		public string strOnRP;
	}
	public struct Sequence
	{
		public Point destination;
		public int speed;
		public int pause;

	}

	public struct Point
	{
		public double row;
		public double col;
	
		public Point(double p_row, double p_col)
		{
			row = Convert.ToDouble(Math.Round(Convert.ToDecimal(p_row),2));
			col = Convert.ToDouble(Math.Round(Convert.ToDecimal(p_col),2));
		}

	}//end of struct

	public struct ResultPacket
	{
        public string strRequestID;
		public string strType;
		public string strTimeStampInSender;
		public string strRouteCacheUsed;
		public string strNumOfRoutesInCache;
	    public string strComType;
		public string strTimeStampInReceiver;
		public string strNotTransmittedDP;
	}

	public class MyTimer2 
	{
		//properties	
		private Thread thread;
		private int _interval;
		private bool _enabled;
		private int _count;
		private int _condition;
		private AutoDataCollection _pc;
		private string _ID;
		public Packet _packet;
		private int _numOfCall;
		public string NTCurrentThread;

		public MyTimer2()
		{
			_count = 0;
			_numOfCall = 0;
			_interval = 100;
			_enabled = false;	
			_condition = -1;				
		}

		public MyTimer2(int p_interval, bool p_enabled)
		{
			_count = 0;
			_numOfCall = 0;
			_interval = p_interval;
			_enabled = p_enabled;
			_condition = -1;
		}

		public MyTimer2(int p_interval, bool p_enabled, AutoDataCollection p_pc)
		{
			_interval = p_interval;
			_enabled = p_enabled;
			_pc = p_pc;
		}
		public MyTimer2(int p_interval, bool p_enabled, int p_condition, 
			AutoDataCollection p_pc, Packet p_packet, string p_ID)
		{
			_count = 0;
			_numOfCall = 0;
			_interval = p_interval;
			_enabled = p_enabled;
			_condition = p_condition;
			_pc = p_pc;
			_packet = p_packet;
			_ID = p_ID;
		}

		public void setEnabled(bool p_enabled)
		{
			_enabled = p_enabled;				 
		}

		public void setInterval(int p_interval)
		{
			_interval = p_interval;
		}

		public int getInterval()
		{
			return _interval;
		}

		public bool getEnabled()
		{
			return _enabled;
		}


		public int getCount()
		{
			return _count;
		}
		public void setCondition(int p_condition)
		{
			_condition = p_condition;
		}

		public void resetCount()
		{
			_count = 0;
		}

		private void RRTimer2_Click()
		{
			try
			{				
				while(_enabled || _pc.send_buffer.Contains(_packet.strRequestID))
				{
					Monitor.Enter(this);
					Thread.Sleep(_interval);
					_count++;
					if(_enabled)
					{
						if(_count<_condition)
						{
							if(_packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
								_packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET) ||
								_packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
							{
								if(_pc.send_buffer.Contains(_packet.strRequestID))
								{
									_pc.SendPacket(ref _packet);
								}
							}
							else //route discovery packet
							{								
								string lostPacketRequestID = _packet.strPayload;
								if(_pc.send_buffer.Contains(lostPacketRequestID) && 
									_pc.send_buffer.Contains(_packet.strRequestID))
								{
                                    _pc.SendPacket(ref _packet);
								}
							}
						}
						else 
						{
							_enabled = false;
							resetCount();
							_pc.ReleaseRRTimer(_packet.strRequestID);
							if(_packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
								_packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
							{
								if(_pc.send_buffer.Contains(_packet.strRequestID))
								{
									_pc.routeCache.RemovePath(_packet.strRouteCache);
									string temp = Packet_CONSTANTS.NOTAVAILABLE;
									_pc.RepairRouteError(temp,ref _packet);
								}
							}
							else //Start Chat Packet, Route Discovery
							{
								if(_packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
								{
									if(_pc.send_buffer.Contains(_packet.strRequestID))
									{
										_pc.send_buffer.Remove(_packet.strRequestID);//SCP packet
										_pc.SendStartChatPacket(ref _packet);
									}
								}
								else//Route Discovery Packet
								{
									if(_pc.send_buffer.Contains(_packet.strRequestID))
									{
										string lostPacketRequestID = _packet.strPayload;
										if(_pc.send_buffer.Contains(lostPacketRequestID))
										{
											_pc.ContinueTryingAfterThreeRDPs(_packet);
										}
									}
								}
							}
							break;
						}
					}//check again
					Monitor.Exit(this);
				}//while
				
			}
			catch(Exception e)
			{
				MessageBox.Show("RRTimer2_Click(): "+e.Message);
			}
		}

		public void SetRRTimer2()
		{
			try
			{
				thread = new Thread(new ThreadStart(this.RRTimer2_Click));
				thread.Start();
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SetRRTimer2(): "+e.Message);
			}
		}

		public void SetReceiptTimer2()
		{			
			thread = new Thread(new ThreadStart(this.ReceiptTimer2_Click));
			thread.Start();
		}

		private void MakeCurrentThreadSleep()
		{
			int count = 0;
			while(500*count <= _interval)
			{
				count++;
				Thread.Sleep(500); 
			}
		}
		private void ReceiptTimer2_Click()
		{
			try
			{
				_numOfCall++;
				while(_enabled)
				{				
					Monitor.Enter(this);
					Thread.Sleep(_interval);
					if((_numOfCall == 1) && (_enabled))
					{
						_pc.ReleaseReceiptTimer(_ID);
						if(_pc.ReceiptBuffer.Contains(_ID))
						{
							_pc.InformRouteErrorToSource(_packet);
						}
						break;
					}												
					Monitor.Exit(this);
				}
				_numOfCall = 0;
			}
			catch(Exception e)
			{
				MessageBox.Show("ReceiptTimer2_Click(): "+e.Message);
			}
		}

	}//inner class


}
